from getpaid.processor import BaseProcessor


class Plugin(BaseProcessor):
    display_name = 'Test plugin'
    accepted_currencies = ['EUR', 'USD']
    slug = 'test_plugin'

    def prepare_transaction(self, *args, **kwargs):
        return ''



import pytest
from django.conf import settings
from django.http import HttpResponse
from getpaid_core.registry import registry as core_registry

from getpaid import FraudStatus, PaymentStatus
from getpaid.processor import BaseProcessor
from getpaid.registry import registry

from .tools import Plugin

dummy = 'getpaid.backends.dummy'


class CoreOnlyPlugin(BaseProcessor):
    display_name = 'Core-only plugin'
    accepted_currencies = ['EUR']
    slug = 'core_only_plugin'

    def prepare_transaction(self, *args, **kwargs):
        return HttpResponse(b'OK')


class TestRegistry:
    @pytest.fixture(autouse=True)
    def setup_plugin(self):
        if Plugin.slug not in registry:
            registry.register(Plugin)

    def test_register(self):
        assert dummy in settings.INSTALLED_APPS
        assert dummy in registry
        assert issubclass(registry[dummy], BaseProcessor)
        assert Plugin.slug in registry

    def test_get_choices(self):
        choices = registry.get_choices('USD')
        assert len(choices) >= 1
        slugs = [c[0] for c in choices]
        assert Plugin.slug in slugs

    def test_url(self):
        assert len(registry.urls) >= 0

    def test_get_all_supported_currency_choices(self):
        choices = registry.get_all_supported_currency_choices()
        currency_codes = {code for code, _label in choices}
        assert 'EUR' in currency_codes
        assert 'USD' in currency_codes
        for code, label in choices:
            assert code == label
            assert len(code) == 3
            assert code == code.upper()

    def test_choices(self):
        fraud_choices = FraudStatus.CHOICES
        assert type(fraud_choices) == tuple

        payment_choices = PaymentStatus.CHOICES
        assert type(payment_choices) == tuple

    def test_module_registry_wraps_core_singleton(self):
        assert registry._core is core_registry

    def test_registering_via_core_is_visible_in_django_registry(self):
        core_registry.register(CoreOnlyPlugin)

        try:
            assert CoreOnlyPlugin.slug in registry
            assert registry[CoreOnlyPlugin.slug] is CoreOnlyPlugin
        finally:
            core_registry.unregister(CoreOnlyPlugin.slug)



"""Django-specific payment processor base class.

Extends getpaid-core's BaseProcessor with Django settings integration,
template handling, and URL helpers.
"""

from collections.abc import Mapping, Sequence
from importlib import import_module
from typing import Any

from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import ImproperlyConfigured
from django.forms import BaseForm
from django.http import HttpRequest
from django.views import View
from getpaid_core.processor import BaseProcessor as CoreBaseProcessor

from .runtime import prepare_payment


class BaseProcessor(CoreBaseProcessor):
    """Django adapter for core BaseProcessor.

    Adds Django settings integration, template handling,
    and URL helpers on top of core's abstract methods.
    """

    production_url: str | None = None
    sandbox_url: str | None = None
    post_form_class: type | None = None
    post_template_name: str | None = None
    client_class: type | str | None = None
    client: object | None = None
    ok_statuses: list[int] = [200]

    def __init__(self, payment, config=None) -> None:
        # Read config from Django settings if not provided
        if config is None:
            path = getattr(payment, 'backend', '') or ''
            config = getattr(settings, 'GETPAID_BACKEND_SETTINGS', {}).get(
                path, {}
            )
        super().__init__(payment, config=config)
        self.path = getattr(payment, 'backend', '') or ''
        self.context: dict = {}
        if self.slug is None:
            self.slug = self.path  # ty: ignore[invalid-attribute-access]
        self.optional_config = getattr(settings, 'GETPAID', {})
        if self.client_class is not None:
            self.client = self.get_client()

    def get_setting(self, name: str, default: Any | None = None) -> Any:
        """Read setting from backend config, falling back to GETPAID global."""
        value = self.config.get(name, default)
        if value is None:
            value = self.optional_config.get(name, None)
        return value

    def get_client_class(self) -> type | None:
        class_path = self.get_setting('CLIENT_CLASS')
        if not class_path:
            class_path = self.client_class
        if class_path and not callable(class_path):
            module_name, _, class_name = class_path.rpartition('.')
            module = import_module(module_name)
            return getattr(module, class_name)
        return class_path  # ty: ignore[invalid-return-type]

    def get_client(self) -> object:
        client_class = self.get_client_class()
        if client_class is None:
            raise ImproperlyConfigured('No client class configured!')
        return client_class(**self.get_client_params())

    def get_client_params(self) -> dict:
        return {}

    @classmethod
    def class_id(cls, **kwargs) -> str:
        return cls.__module__

    @classmethod
    def get_display_name(cls, **kwargs) -> str:
        return cls.display_name

    @classmethod
    def get_accepted_currencies(cls, **kwargs) -> Sequence[str]:
        return cls.accepted_currencies

    @classmethod
    def get_logo_url(cls, **kwargs) -> str | None:
        return cls.logo_url

    @classmethod
    def get_paywall_baseurl(cls, **kwargs) -> str | None:  # ty: ignore[invalid-method-override]
        if settings.DEBUG:
            return cls.sandbox_url
        return cls.production_url

    @staticmethod
    def get_our_baseurl(request: HttpRequest | None = None, **kwargs) -> str:
        """Get base URL for our site.

        Uses Sites framework when no request is available.
        """
        scheme = 'http' if settings.DEBUG else 'https'
        if request is not None:
            domain = get_current_site(request).domain
        else:
            # Circular import workaround: Site model loaded at runtime
            from django.contrib.sites.models import Site

            domain = Site.objects.get_current().domain
        return f'{scheme}://{domain}/'

    def get_template_names(
        self, view: View | None = None, **kwargs
    ) -> list[str]:
        template_name = self.get_setting('POST_TEMPLATE')
        if template_name is None:
            template_name = self.post_template_name
        if (
            template_name is None
            and view is not None
            and hasattr(view, 'get_template_names')
        ):
            return view.get_template_names()  # ty: ignore[call-non-callable]
        if template_name is None:
            raise ImproperlyConfigured("Couldn't determine template name!")
        return [template_name]

    def get_form_class(self, **kwargs) -> type | None:
        form_class_path = self.get_setting('POST_FORM_CLASS')
        if not form_class_path:
            return self.post_form_class
        if isinstance(form_class_path, str):
            module_path, class_name = form_class_path.rsplit('.', 1)
            module = import_module(module_path)
            return getattr(module, class_name)
        return self.post_form_class

    def prepare_form_data(self, post_data: dict, **kwargs) -> Mapping[str, Any]:
        return post_data

    def get_form(self, post_data: dict, **kwargs) -> BaseForm:
        form_class = self.get_form_class()
        if form_class is None:
            raise ImproperlyConfigured("Couldn't determine form class!")
        form_data = self.prepare_form_data(post_data)
        return form_class(fields=form_data)

    def verify_callback(  # ty: ignore[invalid-method-override]
        self, request: HttpRequest, **kwargs
    ) -> None:
        """Verify callback from Django request. Override in backends.

        Default: no-op (accepts all callbacks).
        The signature intentionally differs from the core's async version —
        Django adapter wraps the HTTP request differently.
        """

    def handle_paywall_callback(self, request, **kwargs):
        """Handle paywall callback. Override in backends."""
        raise NotImplementedError

    def fetch_payment_status(self, **kwargs):
        """Fetch payment status from gateway. Override in backends."""
        raise NotImplementedError

    def prepare_transaction(self, request=None, view=None, **kwargs):
        return prepare_payment(
            self.payment,
            request=request,
            view=view,
            **kwargs,
        )



from decimal import Decimal

from asgiref.sync import sync_to_async
from django.utils import timezone


class DjangoPaymentRepository:
    def __init__(self, model_class) -> None:
        self.model_class = model_class

    async def get_by_id(self, payment_id):
        return await sync_to_async(
            self._get_by_id,
            thread_sensitive=True,
        )(payment_id)

    async def create(self, **kwargs):
        return await sync_to_async(
            self._create,
            thread_sensitive=True,
        )(**kwargs)

    async def save(self, payment):
        return await sync_to_async(
            self._save,
            thread_sensitive=True,
        )(payment)

    async def update_status(self, payment_id, status, **fields):
        return await sync_to_async(
            self._update_status,
            thread_sensitive=True,
        )(payment_id, status, **fields)

    async def list_by_order(self, order_id):
        return await sync_to_async(
            self._list_by_order,
            thread_sensitive=True,
        )(order_id)

    def _get_by_id(self, payment_id):
        payment = self.model_class.objects.select_related('order').get(
            pk=payment_id
        )
        return self._normalize_payment(payment)

    def _create(self, **kwargs):
        payment = self.model_class.objects.create(**kwargs)
        return self._normalize_payment(payment)

    def _save(self, payment):
        current_time = timezone.now()
        if (
            getattr(payment, 'amount_paid', 0)
            and getattr(
                payment,
                'last_payment_on',
                None,
            )
            is None
        ):
            payment.last_payment_on = current_time
        if (
            getattr(payment, 'amount_refunded', 0)
            and getattr(
                payment,
                'refunded_on',
                None,
            )
            is None
        ):
            payment.refunded_on = current_time
        payment.save()
        return self._normalize_payment(payment)

    def _update_status(self, payment_id, status, **fields):
        payment = self.model_class.objects.select_related('order').get(
            pk=payment_id
        )
        payment.status = status
        for key, value in fields.items():
            setattr(payment, key, value)
        return self._save(payment)

    def _list_by_order(self, order_id):
        payments = list(
            self.model_class.objects.select_related('order').filter(
                order_id=order_id,
            )
        )
        return [self._normalize_payment(payment) for payment in payments]

    def _normalize_payment(self, payment):
        for field_name in (
            'amount_required',
            'amount_paid',
            'amount_locked',
            'amount_refunded',
        ):
            value = getattr(payment, field_name)
            if not isinstance(value, Decimal):
                setattr(payment, field_name, Decimal(str(value)))
        if payment.provider_data is None:
            payment.provider_data = {}
        return payment



# Generated by Django 3.0.5 on 2020-04-17 21:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('orders', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='custompayment',
            name='fraud_status',
            field=models.CharField(
                choices=[
                    ('unknown', 'unknown'),
                    ('accepted', 'accepted'),
                    ('rejected', 'rejected'),
                    ('check', 'needs manual verification'),
                ],
                db_index=True,
                default='unknown',
                max_length=20,
                verbose_name='fraud status',
            ),
        ),
        migrations.AlterField(
            model_name='custompayment',
            name='status',
            field=models.CharField(
                choices=[
                    ('new', 'new'),
                    ('prepared', 'in progress'),
                    ('pre-auth', 'pre-authed'),
                    ('charge_started', 'charge process started'),
                    ('partially_paid', 'partially paid'),
                    ('paid', 'paid'),
                    ('failed', 'failed'),
                    ('refund_started', 'refund started'),
                    ('refunded', 'refunded'),
                ],
                db_index=True,
                default='new',
                max_length=50,
                verbose_name='status',
            ),
        ),
    ]



from importlib import import_module

from django.conf import settings


def run_getpaid_validators(data: dict) -> dict:
    backend = data['backend']
    getpaid_settings = getattr(settings, 'GETPAID', {})
    global_validators = getpaid_settings.get('VALIDATORS', [])
    backend_settings = getattr(settings, 'GETPAID_BACKEND_SETTINGS', {})
    backend_validators = []

    backend_candidates = [backend]
    if '.' not in backend:
        backend_candidates.append(f'getpaid.backends.{backend}')

    for candidate in backend_candidates:
        validators = backend_settings.get(candidate, {}).get('VALIDATORS', [])
        if validators:
            backend_validators = validators
            break

    unique_validators = list(
        dict.fromkeys(global_validators + backend_validators)
    )
    for path in unique_validators:
        module_name, validator_name = path.rsplit('.', 1)
        module = import_module(module_name)
        validator = getattr(module, validator_name)
        data = validator(data)
    return data



from django.contrib import admin
from django.urls import include, path, re_path
from orders.views import HomeView, OrderView

app_name = 'getpaid_example'

urlpatterns = [
    path('admin/', admin.site.urls),
    re_path(r'^$', HomeView.as_view(), name='home'),
    path('order/<int:pk>/', OrderView.as_view(), name='order_detail'),
    path('payments/', include('getpaid.urls')),
    path('paywall/', include('paywall.urls')),
]



import pytest
from django.conf import settings
from django.contrib.auth.models import User
from django.test import override_settings

pytestmark = pytest.mark.django_db


@override_settings(
    GETPAID_BACKEND_SETTINGS={
        'getpaid.backends.dummy': {'paywall_method': 'REST'}
    }
)
def test_create_payment_view_rejects_anonymous_post(client, order_factory):
    order = order_factory()

    response = client.post(
        '/payments/new/',
        data={
            'order': order.pk,
            'amount_required': '1.00',
            'description': 'Tampered description',
            'currency': 'USD',
            'backend': 'dummy',
        },
    )

    assert response.status_code == 302
    assert settings.LOGIN_URL in response.url


@override_settings(
    GETPAID_BACKEND_SETTINGS={
        'getpaid.backends.dummy': {'paywall_method': 'REST'}
    }
)
def test_create_payment_view_allows_authenticated_post(client, order_factory):
    order = order_factory()
    user = User.objects.create(username='tester')
    client.force_login(user)

    response = client.post(
        '/payments/new/',
        data={
            'order': order.pk,
            'amount_required': '1.00',
            'description': 'Tampered description',
            'currency': 'USD',
            'backend': 'dummy',
        },
    )

    assert response.status_code == 302



import inspect

from asgiref.sync import async_to_sync
from django.core.exceptions import ImproperlyConfigured
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
)
from django.template.response import TemplateResponse
from getpaid_core.enums import BackendMethod, PaymentEvent
from getpaid_core.fsm import apply_payment_update
from getpaid_core.processor import BaseProcessor as CoreBaseProcessor
from getpaid_core.types import PaymentUpdate

from getpaid.adapters import (
    adapt_callback_request,
    call_processor_verify_callback,
)
from getpaid.repository import DjangoPaymentRepository


def prepare_payment(payment, request=None, view=None, **kwargs):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    result = _call_processor_method(
        processor.prepare_transaction,
        request=request,
        view=view,
        **kwargs,
    )
    if isinstance(result, HttpResponse):
        return result
    apply_payment_update(
        payment,
        PaymentUpdate(
            payment_event=PaymentEvent.PREPARED,
            external_id=result.external_id,
            provider_data=result.provider_data,
        ),
    )
    _save_payment(payment)
    return build_payment_response(payment, result, request=request, view=view)


def handle_callback_request(payment, request, **kwargs):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    if _uses_semantic_callback_contract(processor):
        data, headers, raw_body = adapt_callback_request(request)
        if _uses_semantic_verify_contract(processor):
            _call_processor_method(
                processor.verify_callback,
                data,
                headers,
                raw_body=raw_body,
                **kwargs,
            )
        else:
            call_processor_verify_callback(processor, request)
        update = _call_processor_method(
            processor.handle_callback,
            data,
            headers,
            raw_body=raw_body,
            **kwargs,
        )
        if isinstance(update, HttpResponse):
            return update
        apply_payment_update(payment, update)
        _save_payment(payment)
        return HttpResponse(b'OK')

    call_processor_verify_callback(processor, request)
    return processor.handle_paywall_callback(request, **kwargs)


def fetch_payment_status(payment):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    return _call_processor_method(processor.fetch_payment_status)


def fetch_and_update_payment_status(payment):
    payment = _normalize_payment(payment)
    update = fetch_payment_status(payment)
    apply_payment_update(payment, update)
    _save_payment(payment)
    return payment


def charge_payment(payment, amount=None, **kwargs):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    result = _call_processor_method(processor.charge, amount=amount, **kwargs)
    if result.success:
        if result.async_call:
            update = PaymentUpdate(
                payment_event=PaymentEvent.CHARGE_REQUESTED,
                provider_data=result.provider_data,
            )
        else:
            update = PaymentUpdate(
                payment_event=PaymentEvent.PAYMENT_CAPTURED,
                paid_amount=payment.amount_paid + result.amount_charged,
                provider_data=result.provider_data,
            )
        apply_payment_update(payment, update)
        _save_payment(payment)
    return result


def release_payment_lock(payment, **kwargs):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    amount = _call_processor_method(processor.release_lock, **kwargs)
    apply_payment_update(
        payment,
        PaymentUpdate(payment_event=PaymentEvent.LOCK_RELEASED),
    )
    _save_payment(payment)
    return amount


def start_payment_refund(payment, amount=None, **kwargs):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    result = _call_processor_method(
        processor.start_refund,
        amount=amount,
        **kwargs,
    )
    apply_payment_update(
        payment,
        PaymentUpdate(
            payment_event=PaymentEvent.REFUND_REQUESTED,
            provider_data=result.provider_data,
        ),
    )
    _save_payment(payment)
    return result


def cancel_payment_refund(payment, **kwargs):
    payment = _normalize_payment(payment)
    processor = payment._get_processor()
    success = _call_processor_method(processor.cancel_refund, **kwargs)
    if success:
        apply_payment_update(
            payment,
            PaymentUpdate(payment_event=PaymentEvent.REFUND_CANCELLED),
        )
        _save_payment(payment)
    return success


def build_payment_response(payment, result, request=None, view=None):
    if result.method is BackendMethod.POST:
        processor = payment._get_processor()
        if not hasattr(processor, 'get_form') or not hasattr(
            processor,
            'get_template_names',
        ):
            raise ImproperlyConfigured(
                'POST-based payments require a Django-aware processor.'
            )
        form = processor.get_form(result.form_data or {})
        return TemplateResponse(
            request=request,
            template=processor.get_template_names(view=view),
            context={
                'form': form,
                'paywall_url': result.redirect_url or '#',
            },
        )
    redirect_url = (
        result.redirect_url
        or payment._get_processor().get_our_baseurl(
            request,
        )
    )
    return HttpResponseRedirect(redirect_url)


def _call_processor_method(method, *args, **kwargs):
    if inspect.iscoroutinefunction(method):
        return async_to_sync(method)(*args, **kwargs)
    return method(*args, **kwargs)


def _save_payment(payment):
    repository = DjangoPaymentRepository(type(payment))
    return repository._save(payment)


def _normalize_payment(payment):
    repository = DjangoPaymentRepository(type(payment))
    return repository._normalize_payment(payment)


def _uses_semantic_callback_contract(processor):
    handle_method = getattr(processor, 'handle_callback', None)
    if handle_method is None:
        return False
    type_handle_method = getattr(type(processor), 'handle_callback', None)
    if type_handle_method is None:
        return inspect.iscoroutinefunction(handle_method)
    return type_handle_method is not CoreBaseProcessor.handle_callback


def _uses_semantic_verify_contract(processor):
    verify_method = getattr(processor, 'verify_callback', None)
    if verify_method is None:
        return False
    return inspect.iscoroutinefunction(verify_method)



"""Exception hierarchy -- re-exports from getpaid-core."""

from getpaid_core.exceptions import (
    ChargeFailure,
    CommunicationError,
    CredentialsError,
    GetPaidException,
    InvalidCallbackError,
    InvalidTransitionError,
    LockFailure,
    RefundFailure,
)

__all__ = [
    'ChargeFailure',
    'CommunicationError',
    'CredentialsError',
    'GetPaidException',
    'InvalidCallbackError',
    'InvalidTransitionError',
    'LockFailure',
    'RefundFailure',
]



import factory
import swapper
from paywall.models import PaymentEntry


class OrderFactory(factory.django.DjangoModelFactory):
    name = factory.Faker('color_name')
    total = factory.Faker(
        'pydecimal', positive=True, right_digits=2, min_value=10, max_value=500
    )
    currency = 'EUR'

    class Meta:
        model = swapper.load_model('getpaid', 'Order')


class PaymentFactory(factory.django.DjangoModelFactory):
    order = factory.SubFactory(OrderFactory)
    amount_required = factory.SelfAttribute('order.total')
    currency = factory.SelfAttribute('order.currency')
    description = factory.SelfAttribute('order.name')
    backend = 'getpaid.backends.dummy'

    class Meta:
        model = swapper.load_model('getpaid', 'Payment')


class PaywallEntryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = PaymentEntry



"""Sphinx configuration for django-getpaid."""

import os
import sys
from pathlib import Path

import django

# Add project and example app to path for autodoc
sys.path.insert(0, str(Path('..').resolve()))
sys.path.insert(0, str(Path('../example').resolve()))
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'example.settings')
django.setup()

import getpaid

project = 'django-getpaid'
author = 'Dominik Kozaczko'
project_copyright = '2012-2013 Krzysztof Dorosz, 2013-2026 Dominik Kozaczko'

version = '.'.join(getpaid.__version__.split('.')[:2])
release = getpaid.__version__

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosectionlabel',
    'myst_parser',
]

autodoc_typehints = 'description'
autodoc_member_order = 'bysource'

html_theme = 'furo'
html_title = 'django-getpaid'

source_suffix = {
    '.md': 'markdown',
}

myst_enable_extensions = [
    'colon_fence',
    'fieldlist',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    # Django and getpaid-core intersphinx will work on ReadTheDocs.
    # Uncomment when building with network access:
    # 'django': (
    #     'https://docs.djangoproject.com/en/5.1/',
    #     'https://docs.djangoproject.com/en/5.1/_objects.inv',
    # ),
    # 'getpaid-core': ('https://getpaid-core.readthedocs.io/en/latest/', None),
}

# Suppress duplicate label warnings from autosectionlabel
# and intersphinx connection failures (work fine on ReadTheDocs)
suppress_warnings = ['autosectionlabel.*', 'intersphinx.external']



import json

import httpx
from django.conf import settings
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import FormView

from getpaid.status import PaymentStatus as ps

from .forms import QuestionForm
from .models import PaymentEntry


class AuthorizationView(FormView):
    """
    This view simulates the behavior of payment broker
    """

    form_class = QuestionForm
    template_name = 'paywall/fake_gateway_authorization_form.html'
    success = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        params = (
            self.request.POST or self.request.GET
        )  # both cases for testability
        if 'pay_id' in params:
            obj = get_object_or_404(PaymentEntry, id=params.get('pay_id'))
            context['ext_id'] = obj.ext_id
            context['value'] = obj.value
            context['currency'] = obj.currency
            context['description'] = obj.description

            context['message'] = 'Presenting pre-registered payment'
        else:
            context['ext_id'] = params.get('ext_id')
            context['value'] = params.get('value')
            context['currency'] = params.get('currency')
            context['description'] = params.get('description')

            context['message'] = 'Presenting directly requested payment'
        return context

    def get_initial(self):
        initial = super().get_initial()
        params = self.request.POST or self.request.GET
        if 'pay_id' in params:
            obj = get_object_or_404(PaymentEntry, id=params.get('pay_id'))
            initial['callback'] = obj.callback
            initial['success_url'] = obj.success_url
            initial['failure_url'] = obj.failure_url
        else:
            initial['callback'] = params.get('callback', '')
            initial['success_url'] = params.get('success_url', '')
            initial['failure_url'] = params.get('failure_url', '')
        return initial

    def get_success_url(self):
        if self.success:
            return self.form.cleaned_data['success_url']
        return self.form.cleaned_data['failure_url']

    def form_valid(self, form):
        self.form = form
        callback = form.cleaned_data['callback']
        url = self.request.build_absolute_uri(callback) if callback else None
        if url:
            if form.cleaned_data['authorize_payment'] == '1':
                self.success = True
                if settings.PAYWALL_MODE == 'LOCK':
                    httpx.post(url, json={'new_status': ps.PRE_AUTH})
                else:
                    httpx.post(url, json={'new_status': ps.PAID})
            else:
                self.success = False
                httpx.post(url, json={'new_status': ps.FAILED})
        return super().form_valid(form)


authorization_view = csrf_exempt(AuthorizationView.as_view())


def get_status(request, pk, **kwargs):
    obj = get_object_or_404(PaymentEntry, pk=pk)
    return JsonResponse({
        'payment_status': obj.payment_status,
        'fraud_status': obj.fraud_status,
    })


@csrf_exempt
def rest_register_payment(request):
    legal_fields = [
        'ext_id',
        'value',
        'currency',
        'description',
        'callback',
        'success_url',
        'failure_url',
    ]
    params = {
        k: v for k, v in json.loads(request.body).items() if k in legal_fields
    }
    payment = PaymentEntry.objects.create(**params)

    url = request.build_absolute_uri(reverse('paywall:gateway'))
    url += f'?pay_id={payment.id}'

    content = {'url': url}
    return JsonResponse(content)


@csrf_exempt
def rest_operation(request):
    """
    For test purposes backend can "suggest" the flow of payment.
    """
    data = json.loads(request.body)

    payment_id = data['id']
    new_status = data['new_status']

    payment = PaymentEntry.objects.create(id=payment_id)
    if new_status == ps.PRE_AUTH:
        payment.send_confirm_lock()
    elif new_status == ps.PAID:
        if payment.payment_status == ps.PRE_AUTH:
            payment.send_confirm_charge()
        elif payment.payment_status == ps.REFUND_STARTED:
            payment.cancel_refund()
        else:
            raise NotImplementedError(
                f'Cannot handle change to {ps.PAID} from {payment.payment_status}'
            )
    elif new_status == ps.FAILED:
        payment.send_fail()
    elif new_status == ps.REFUND_STARTED:
        payment.start_refund()
    elif new_status == ps.REFUNDED:
        payment.send_confirm_refund()

    return JsonResponse({'id': str(payment_id), 'new_status': new_status})



from django.db import models
from django.urls import reverse

from getpaid.abstracts import AbstractOrder, AbstractPayment

ORDER_STATUS_CHOICES = (('W', 'Waiting for payment'), ('P', 'Payment complete'))


class Order(AbstractOrder):
    """
    This is an example Order object. This one is very simple - only one item,
    but you can easily create more complicated models with multi-items
    as it does not matter for payment processing.
    """

    name = models.CharField(
        max_length=100, default='Lock, Stock and Two Smoking Barrels'
    )
    total = models.DecimalField(
        decimal_places=2, max_digits=8, default='199.99'
    )
    currency = models.CharField(max_length=3, default='EUR')
    status = models.CharField(
        max_length=1, blank=True, default='W', choices=ORDER_STATUS_CHOICES
    )

    def get_absolute_url(self):
        return reverse('order_detail', kwargs={'pk': self.pk})

    def __str__(self):
        return self.name

    def clean(self):
        self.currency = self.currency.upper()

    def get_total_amount(self):
        return self.total

    def get_description(self):
        return self.name

    def get_buyer_info(self):
        return {'email': 'test@example.com'}


class CustomPayment(AbstractPayment):
    """
    Needed for test purposes.
    """

    custom = models.BooleanField(default=True, editable=False)



from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ('orders', '0005_remove_fsm_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='custompayment',
            name='provider_data',
            field=models.JSONField(
                blank=True,
                default=dict,
                verbose_name='provider data',
            ),
        ),
    ]



"""Payment types and enums.

Django-specific enum wrappers with choices property.
Values are kept identical to getpaid-core and verified at import time.
Re-exports core types unchanged.
"""

from decimal import Decimal
from enum import Enum
from typing import Any, TypedDict

from django.http import HttpResponse
from django.utils.functional import classproperty
from django.utils.translation import pgettext_lazy
from getpaid_core.enums import (
    BackendMethod,
    ConfirmationMethod,
)
from getpaid_core.enums import FraudStatus as _CoreFraudStatus
from getpaid_core.enums import PaymentStatus as _CorePaymentStatus
from getpaid_core.types import BuyerInfo, ChargeResponse, ItemInfo

# Re-export core types directly
__all__ = [
    'BackendMethod',
    'BuyerInfo',
    'ChargeResponse',
    'ConfirmationMethod',
    'FraudStatus',
    'ItemInfo',
    'PaymentStatus',
    'PaymentStatusResponse',
    'RestfulResult',
]


class PaymentStatus(str, Enum):
    """Payment status with Django choices support.

    Values are identical to getpaid-core's PaymentStatus.
    Adds .choices and .CHOICES classproperties for Django model fields.
    """

    NEW = 'new'
    PREPARED = 'prepared'
    PRE_AUTH = 'pre-auth'
    IN_CHARGE = 'charge_started'
    PARTIAL = 'partially_paid'
    PAID = 'paid'
    FAILED = 'failed'
    REFUND_STARTED = 'refund_started'
    REFUNDED = 'refunded'

    @classproperty
    def choices(cls):
        return (
            (cls.NEW.value, pgettext_lazy('payment status', 'new')),  # ty: ignore[possibly-missing-attribute]
            (
                cls.PREPARED.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'in progress'),
            ),
            (
                cls.PRE_AUTH.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'pre-authed'),
            ),
            (
                cls.IN_CHARGE.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'charge process started'),
            ),
            (
                cls.PARTIAL.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'partially paid'),
            ),
            (cls.PAID.value, pgettext_lazy('payment status', 'paid')),  # ty: ignore[possibly-missing-attribute]
            (
                cls.FAILED.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'failed'),
            ),
            (
                cls.REFUND_STARTED.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'refund started'),
            ),
            (
                cls.REFUNDED.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('payment status', 'refunded'),
            ),
        )

    @classproperty
    def CHOICES(cls):
        """Backward compatibility."""
        return cls.choices


class FraudStatus(str, Enum):
    """Fraud status with Django choices support.

    Values are identical to getpaid-core's FraudStatus.
    Adds .choices and .CHOICES classproperties for Django model fields.
    """

    UNKNOWN = 'unknown'
    ACCEPTED = 'accepted'
    REJECTED = 'rejected'
    CHECK = 'check'

    @classproperty
    def choices(cls):
        return (
            (
                cls.UNKNOWN.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('fraud status', 'unknown'),
            ),
            (
                cls.ACCEPTED.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('fraud status', 'accepted'),
            ),
            (
                cls.REJECTED.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('fraud status', 'rejected'),
            ),
            (
                cls.CHECK.value,  # ty: ignore[possibly-missing-attribute]
                pgettext_lazy('fraud status', 'needs manual verification'),
            ),
        )

    @classproperty
    def CHOICES(cls):
        """Backward compatibility."""
        return cls.choices


# Runtime verification that values match core enums
def _verify_enum_sync():
    for member in _CorePaymentStatus:
        local = PaymentStatus[member.name]
        if local.value != member.value:
            msg = (
                f'PaymentStatus.{member.name} value mismatch: '
                f'{local.value!r} != {member.value!r}'
            )
            raise AssertionError(msg)
    for member in _CoreFraudStatus:
        local = FraudStatus[member.name]
        if local.value != member.value:
            msg = (
                f'FraudStatus.{member.name} value mismatch: '
                f'{local.value!r} != {member.value!r}'
            )
            raise AssertionError(msg)


_verify_enum_sync()


class GetpaidInternalResponse(TypedDict):
    raw_response: Any
    exception: Exception | None


class PaymentStatusResponse(GetpaidInternalResponse):
    amount: Decimal | None
    callback: str | None
    callback_result: Any | None
    saved: bool | None


class FormField(TypedDict):
    name: str
    value: Any
    label: str | None
    widget: str
    help_text: str | None
    required: bool


class PaymentForm(TypedDict):
    fields: list[FormField]


class RestfulResult(TypedDict):
    status_code: int
    result: HttpResponse
    target_url: str | None
    form: PaymentForm | None
    message: str | bytes | None



from decimal import Decimal

from django.http import HttpResponseBadRequest
from getpaid_core.backends.dummy import DummyProcessor
from getpaid_core.enums import BackendMethod, PaymentEvent
from getpaid_core.types import (
    ChargeResult,
    PaymentUpdate,
    RefundResult,
    TransactionResult,
)

from getpaid.post_forms import PaymentHiddenInputsPostForm
from getpaid.processor import BaseProcessor


class PaymentProcessor(BaseProcessor, DummyProcessor):
    slug = 'dummy'
    display_name = 'Dummy'
    accepted_currencies = ['PLN', 'EUR', 'USD', 'GBP', 'CHF', 'CZK']
    post_form_class = PaymentHiddenInputsPostForm
    post_template_name = 'getpaid_dummy/payment_post_form.html'
    sandbox_url = 'https://dummy.example.com'
    production_url = 'https://dummy.example.com'

    def get_paywall_method(self):
        return self.get_setting('paywall_method', 'REST')

    async def prepare_transaction(self, **kwargs) -> TransactionResult:
        method = BackendMethod(self.get_paywall_method())
        if method is BackendMethod.POST:
            return TransactionResult(
                method=method,
                redirect_url='https://dummy.example.com/form',
                form_data={
                    'payment_id': str(self.payment.id),
                    'amount': f'{self.payment.amount_required:.2f}',
                    'currency': self.payment.currency,
                },
            )
        return TransactionResult(
            method=method,
            redirect_url=f'https://dummy.example.com/pay/{self.payment.id}',
        )

    async def handle_callback(self, data, headers, **kwargs):
        if 'event' in data:
            return await DummyProcessor.handle_callback(
                self, data, headers, **kwargs
            )

        new_status = data.get('new_status')
        if not new_status:
            return HttpResponseBadRequest(b'Missing new_status')

        if new_status == 'paid':
            return await DummyProcessor.handle_callback(
                self,
                {
                    'event': 'payment_confirmed',
                    'paid_amount': str(self.payment.amount_required),
                },
                headers,
                **kwargs,
            )
        if new_status == 'pre-auth':
            return await DummyProcessor.handle_callback(
                self,
                {
                    'event': 'payment_locked',
                    'locked_amount': str(self.payment.amount_required),
                },
                headers,
                **kwargs,
            )
        if new_status == 'failed':
            return await DummyProcessor.handle_callback(
                self,
                {'event': 'payment_failed'},
                headers,
                **kwargs,
            )
        if new_status == 'refund_started':
            return PaymentUpdate(payment_event=PaymentEvent.REFUND_REQUESTED)
        if new_status == 'refunded':
            return PaymentUpdate(
                payment_event=PaymentEvent.REFUND_CONFIRMED,
                refunded_amount=self.payment.amount_paid,
            )
        return HttpResponseBadRequest(
            f'Unhandled status: {new_status}'.encode()
        )

    def handle_paywall_callback(self, request, **kwargs):
        from getpaid.runtime import handle_callback_request

        return handle_callback_request(self.payment, request, **kwargs)

    async def fetch_payment_status(self, **kwargs):
        confirmation_status = self.get_setting('confirmation_status', 'paid')
        confirmation_event = {
            'paid': 'payment_confirmed',
            'pre_auth': 'payment_locked',
            'failed': 'payment_failed',
        }.get(confirmation_status, 'payment_confirmed')
        self.config['confirmation_event'] = confirmation_event
        return await DummyProcessor.fetch_payment_status(self, **kwargs)

    async def charge(self, amount=None, **kwargs) -> ChargeResult:
        return await DummyProcessor.charge(self, amount=amount, **kwargs)

    async def release_lock(self, **kwargs):
        return await DummyProcessor.release_lock(self, **kwargs)

    async def start_refund(self, amount=None, **kwargs) -> RefundResult:
        refund_amount = (
            amount if amount is not None else self.payment.amount_paid
        )
        return RefundResult(
            amount=Decimal(str(refund_amount)),
            provider_data={'refund_id': f'dummy-refund-{self.payment.id}'},
        )

    async def cancel_refund(self, **kwargs):
        return True



import logging
import uuid
from decimal import Decimal
from importlib import import_module
from typing import cast

import swapper
from django import forms
from django.db import models
from django.db.transaction import atomic
from django.forms import BaseForm
from django.http import HttpRequest, HttpResponse
from django.shortcuts import resolve_url
from django.utils.translation import gettext_lazy as _
from django.views import View
from getpaid_core.enums import FraudEvent
from getpaid_core.fsm import apply_payment_update
from getpaid_core.protocols import Payment as CorePaymentProtocol
from getpaid_core.types import PaymentUpdate

from getpaid.exceptions import ChargeFailure
from getpaid.runtime import (
    cancel_payment_refund,
    charge_payment,
    fetch_and_update_payment_status,
    fetch_payment_status,
    prepare_payment,
    release_payment_lock,
    start_payment_refund,
)
from getpaid.types import (
    BuyerInfo,
    ChargeResponse,
    ItemInfo,
    RestfulResult,
)
from getpaid.types import FraudStatus as fs
from getpaid.types import PaymentStatus as ps

logger = logging.getLogger(__name__)


class AbstractOrder(models.Model):
    """Base class for Order models.

    Consider using UUIDField as primary or secondary key to hide volume.
    """

    class Meta:
        abstract = True

    def get_return_url(
        self,
        *args,
        success: bool | None = None,
        **kwargs,
    ) -> str:
        """Return URL after payment completion.

        Override to customize. Default: get_absolute_url().
        """
        return self.get_absolute_url()

    def get_absolute_url(self) -> str:
        raise NotImplementedError

    def is_ready_for_payment(self) -> bool:
        """Validate order is ready for payment.

        Override for custom validation. Raise ValidationError
        for verbose error messages.
        """
        if self.payments.exclude(status=ps.FAILED).exists():  # ty: ignore[unresolved-attribute]
            raise forms.ValidationError(
                _('Non-failed Payments exist for this Order.')
            )
        return True

    def get_items(self) -> list[ItemInfo]:
        """Return list of items for the payment."""
        return [
            {
                'name': self.get_description(),
                'quantity': 1,
                'unit_price': self.get_total_amount(),
            }
        ]

    def get_total_amount(self) -> Decimal:
        raise NotImplementedError

    def get_currency(self) -> str:
        field_name = 'currency'
        return cast('str', getattr(self, field_name))

    def get_buyer_info(self) -> BuyerInfo:
        raise NotImplementedError

    def get_description(self) -> str:
        raise NotImplementedError


class AbstractPayment(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    order = models.ForeignKey(
        swapper.get_model_name('getpaid', 'Order'),
        verbose_name=_('order'),
        on_delete=models.CASCADE,
        related_name='payments',
    )
    amount_required = models.DecimalField(
        _('amount required'),
        decimal_places=2,
        max_digits=20,
        help_text=_(
            'Amount required to fulfill the payment; '
            'in selected currency, normal notation'
        ),
    )
    currency = models.CharField(_('currency'), max_length=3)
    status = models.CharField(
        _('status'),
        max_length=50,
        choices=ps.choices,
        default=ps.NEW,
        db_index=True,
    )
    backend = models.CharField(_('backend'), max_length=100, db_index=True)
    created_on = models.DateTimeField(
        _('created on'), auto_now_add=True, db_index=True
    )
    last_payment_on = models.DateTimeField(
        _('paid on'),
        blank=True,
        null=True,
        default=None,
        db_index=True,
    )
    amount_locked = models.DecimalField(
        _('amount locked'),
        decimal_places=2,
        max_digits=20,
        default=0,
        help_text=_('Amount locked with this payment, ready to charge.'),
    )
    amount_paid = models.DecimalField(
        _('amount paid'),
        decimal_places=2,
        max_digits=20,
        default=0,
        help_text=_('Amount actually paid.'),
    )
    refunded_on = models.DateTimeField(
        _('refunded on'),
        blank=True,
        null=True,
        default=None,
        db_index=True,
    )
    amount_refunded = models.DecimalField(
        _('amount refunded'),
        decimal_places=2,
        max_digits=20,
        default=0,
    )
    external_id = models.CharField(
        _('external id'),
        max_length=64,
        blank=True,
        db_index=True,
        default='',
    )
    description = models.CharField(
        _('description'), max_length=128, blank=True, default=''
    )
    fraud_status = models.CharField(
        _('fraud status'),
        max_length=20,
        choices=fs.choices,
        default=fs.UNKNOWN,
        db_index=True,
    )
    fraud_message = models.TextField(_('fraud message'), blank=True)
    provider_data = models.JSONField(
        _('provider data'), default=dict, blank=True
    )

    class Meta:
        abstract = True
        ordering = ['-created_on']
        verbose_name = _('Payment')
        verbose_name_plural = _('Payments')

    def __str__(self):
        return f'Payment #{self.id}'

    # ---- Properties ----

    @property
    def fully_paid(self) -> bool:
        return self.amount_paid >= self.amount_required

    def is_fully_paid(self) -> bool:
        return self.amount_paid >= self.amount_required

    def is_fully_refunded(self) -> bool:
        return (
            self.amount_refunded > 0
            and self.amount_refunded >= self.amount_paid
        )

    def flag_as_fraud(self, message=''):
        payment = cast(CorePaymentProtocol, self)
        apply_payment_update(
            payment,
            PaymentUpdate(
                fraud_event=FraudEvent.REJECT,
                fraud_message=message,
            ),
        )

    def flag_as_legit(self, message=''):
        payment = cast(CorePaymentProtocol, self)
        apply_payment_update(
            payment,
            PaymentUpdate(
                fraud_event=FraudEvent.ACCEPT,
                fraud_message=message,
            ),
        )

    def flag_for_check(self, message=''):
        payment = cast(CorePaymentProtocol, self)
        apply_payment_update(
            payment,
            PaymentUpdate(
                fraud_event=FraudEvent.REVIEW,
                fraud_message=message,
            ),
        )

    # ---- Delegation helpers ----

    def get_unique_id(self) -> str:
        """Return unique identifier for this payment."""
        return str(self.id)

    def get_items(self) -> list[ItemInfo]:
        """Relay to order's get_items()."""
        return self.order.get_items()  # ty: ignore[possibly-missing-attribute]

    def get_buyer_info(self) -> BuyerInfo:
        return self.order.get_buyer_info()  # ty: ignore[possibly-missing-attribute]

    def _get_processor(self):
        """Get processor instance for this payment's backend."""
        # Circular import workaround: registry imports processor
        from getpaid.registry import registry

        if self.backend in registry:
            processor_class = registry[self.backend]
        else:
            # last resort if backend removed from INSTALLED_APPS
            module = import_module(str(self.backend))
            processor_class = module.PaymentProcessor
        return processor_class(self)

    def get_form(self, *args, **kwargs) -> BaseForm:
        """Interfaces processor's get_form."""
        return self._get_processor().get_form(*args, **kwargs)

    def get_template_names(self, view: View | None = None) -> list[str]:
        """Interfaces processor's get_template_names."""
        return self._get_processor().get_template_names(view=view)

    def handle_paywall_callback(
        self, request: HttpRequest, **kwargs
    ) -> HttpResponse:
        from getpaid.runtime import handle_callback_request

        return handle_callback_request(self, request, **kwargs)

    def fetch_status(self):
        return fetch_payment_status(self)

    @atomic
    def fetch_and_update_status(self):
        fetch_and_update_payment_status(self)
        return self

    def get_return_redirect_url(
        self, request: HttpRequest, success: bool
    ) -> str:
        """Determine redirect URL after payment."""
        processor = self._get_processor()
        if success:
            url = processor.get_setting('SUCCESS_URL')
        else:
            url = processor.get_setting('FAILURE_URL')

        if url is not None:
            kwargs = self.get_return_redirect_kwargs(request, success)
            return resolve_url(url, **kwargs)
        return resolve_url(self.order.get_return_url(self, success=success))  # ty: ignore[possibly-missing-attribute]

    def get_return_redirect_kwargs(
        self, request: HttpRequest, success: bool
    ) -> dict:
        return {'pk': self.id}

    # ---- Action helpers (delegate to processor) ----

    def prepare_transaction(
        self,
        request: HttpRequest | None = None,
        view: View | None = None,
        **kwargs,
    ) -> HttpResponse:
        return prepare_payment(self, request=request, view=view, **kwargs)

    def prepare_transaction_for_rest(
        self,
        request: HttpRequest | None = None,
        view: View | None = None,
        **kwargs,
    ) -> RestfulResult:
        result = self.prepare_transaction(request=request, view=view, **kwargs)
        data = {'status_code': result.status_code, 'result': result}
        if result.status_code == 200:
            data['target_url'] = result.context_data['paywall_url']  # ty: ignore[unresolved-attribute]
            data['form'] = {
                'fields': [
                    {
                        'name': name,
                        'value': field.initial,
                        'label': field.label or name,
                        'widget': field.widget.__class__.__name__,
                        'help_text': field.help_text,
                        'required': field.required,
                    }
                    for name, field in result.context_data[  # ty: ignore[unresolved-attribute]
                        'form'
                    ].fields.items()
                ],
            }
        elif result.status_code == 302:
            data['target_url'] = result.url  # ty: ignore[unresolved-attribute]
        else:
            data['message'] = result.content
        return data  # ty: ignore[invalid-return-type]

    @atomic
    def charge(
        self,
        amount: Decimal | float | int | None = None,
        **kwargs,
    ) -> ChargeResponse:
        amount_to_charge = amount
        if amount_to_charge is None:
            amount_to_charge = self.amount_locked
        if amount_to_charge > self.amount_locked:
            raise ValueError('Cannot charge more than locked value.')
        result = charge_payment(self, amount=amount_to_charge, **kwargs)
        if not result.success:
            raise ChargeFailure(
                'Error occurred while trying to charge locked amount.'
            )
        return result

    @atomic
    def release_lock(self, **kwargs):
        return release_payment_lock(self, **kwargs)

    @atomic
    def start_refund(
        self,
        amount: Decimal | float | int | None = None,
        **kwargs,
    ):
        refund_amount = Decimal(str(amount)) if amount is not None else None
        return start_payment_refund(self, amount=refund_amount, **kwargs)

    @atomic
    def cancel_refund(self, **kwargs):
        return cancel_payment_refund(self, **kwargs)



from collections import OrderedDict
from decimal import Decimal
from unittest.mock import MagicMock, patch

import pytest
import swapper
from django import forms
from django.http import HttpResponseRedirect

from getpaid.status import FraudStatus as fs
from getpaid.status import PaymentStatus as ps

pytestmark = pytest.mark.django_db

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')


def _make_payment(**kwargs):
    order = Order.objects.create()
    defaults = {
        'order': order,
        'currency': order.currency,
        'amount_required': Decimal(str(order.get_total_amount())),
        'backend': 'getpaid.backends.dummy',
        'description': order.get_description(),
    }
    defaults.update(kwargs)
    return Payment.objects.create(**defaults)


class TestPrepareTransaction:
    def test_prepare_transaction_passes_view_to_processor(self):
        payment = _make_payment()
        mock_view = MagicMock()
        mock_request = MagicMock()

        proc = payment._get_processor()
        with (
            patch.object(
                type(proc),
                'prepare_transaction',
                return_value=HttpResponseRedirect('/redirect'),
            ) as mock_prep,
            patch.object(type(payment), '_get_processor', return_value=proc),
        ):
            payment.prepare_transaction(request=mock_request, view=mock_view)
            mock_prep.assert_called_once_with(
                request=mock_request,
                view=mock_view,
            )

    def test_prepare_transaction_for_rest_iterates_form_fields(self):
        payment = _make_payment()

        mock_form = MagicMock()
        mock_form.fields = OrderedDict([
            ('amount', forms.DecimalField(initial='100.00', label='Amount')),
            ('currency', forms.CharField(initial='EUR', label='Currency')),
        ])

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.context_data = {
            'paywall_url': 'https://example.com/pay',
            'form': mock_form,
        }

        with patch.object(
            type(payment),
            'prepare_transaction',
            return_value=mock_response,
        ):
            result = payment.prepare_transaction_for_rest()

        assert result['status_code'] == 200
        assert result['target_url'] == 'https://example.com/pay'
        assert len(result['form']['fields']) == 2
        assert result['form']['fields'][0]['name'] == 'amount'
        assert result['form']['fields'][1]['name'] == 'currency'


class TestChargeAndRefunds:
    def test_charge_does_not_double_count_amount_paid(self):
        payment = _make_payment(
            status=ps.PRE_AUTH,
            amount_locked=Decimal('100.00'),
        )

        payment.charge(amount=Decimal('100.00'))
        payment.refresh_from_db()

        assert payment.amount_paid == Decimal('100.00')
        assert payment.amount_locked == Decimal('0.00')

    def test_charge_partial_amount(self):
        payment = _make_payment(
            status=ps.PRE_AUTH,
            amount_locked=Decimal('100.00'),
        )

        payment.charge(amount=Decimal('50.00'))
        payment.refresh_from_db()

        assert payment.amount_paid == Decimal('50.00')
        assert payment.amount_locked == Decimal('50.00')
        assert payment.status == ps.PARTIAL

    def test_fetch_and_update_status_returns_payment_instance(self, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'confirmation_status': 'paid'}
        }
        payment = _make_payment(status=ps.PREPARED)

        result = payment.fetch_and_update_status()

        assert result is payment
        assert payment.status == ps.PAID


class TestFraudMethodNames:
    def test_payment_has_flag_as_fraud(self):
        payment = _make_payment()
        assert hasattr(payment, 'flag_as_fraud')
        assert callable(payment.flag_as_fraud)

    def test_payment_has_flag_as_legit(self):
        payment = _make_payment()
        assert hasattr(payment, 'flag_as_legit')
        assert callable(payment.flag_as_legit)

    def test_payment_has_flag_for_check(self):
        payment = _make_payment()
        assert hasattr(payment, 'flag_for_check')
        assert callable(payment.flag_for_check)

    def test_flag_as_fraud_transitions_from_unknown(self):
        payment = _make_payment()
        assert payment.fraud_status == fs.UNKNOWN
        payment.flag_as_fraud(message='Suspicious IP')
        payment.save()
        assert payment.fraud_status == fs.REJECTED
        assert payment.fraud_message == 'Suspicious IP'

    def test_flag_as_legit_transitions_from_unknown(self):
        payment = _make_payment()
        assert payment.fraud_status == fs.UNKNOWN
        payment.flag_as_legit(message='Verified buyer')
        payment.save()
        assert payment.fraud_status == fs.ACCEPTED
        assert payment.fraud_message == 'Verified buyer'

    def test_flag_for_check_transitions_from_unknown(self):
        payment = _make_payment()
        assert payment.fraud_status == fs.UNKNOWN
        payment.flag_for_check(message='Needs review')
        payment.save()
        assert payment.fraud_status == fs.CHECK
        assert payment.fraud_message == 'Needs review'

    def test_no_triple_underscore_methods(self):
        payment = _make_payment()
        assert not hasattr(payment, '___mark_as_fraud')
        assert not hasattr(payment, '___mark_as_legit')
        assert not hasattr(payment, '___mark_for_check')


class TestAmountRefundedPrecision:
    def test_amount_refunded_has_2_decimal_places(self):
        field = Payment._meta.get_field('amount_refunded')
        assert field.decimal_places == 2

    def test_all_money_fields_same_precision(self):
        money_fields = [
            'amount_required',
            'amount_locked',
            'amount_paid',
            'amount_refunded',
        ]
        precisions = {}
        for name in money_fields:
            field = Payment._meta.get_field(name)
            precisions[name] = field.decimal_places
        unique_precisions = set(precisions.values())
        assert len(unique_precisions) == 1


class TestFieldVerboseNames:
    def test_amount_locked_verbose_name(self):
        field = Payment._meta.get_field('amount_locked')
        verbose = str(field.verbose_name).lower()
        assert 'locked' in verbose

    def test_amount_paid_verbose_name(self):
        field = Payment._meta.get_field('amount_paid')
        verbose = str(field.verbose_name).lower()
        assert 'paid' in verbose

    def test_provider_data_field_exists(self):
        field = Payment._meta.get_field('provider_data')
        assert field.default == dict

    def test_no_duplicate_verbose_names_on_money_fields(self):
        money_fields = [
            'amount_required',
            'amount_locked',
            'amount_paid',
            'amount_refunded',
        ]
        names = {}
        for name in money_fields:
            field = Payment._meta.get_field(name)
            names[name] = str(field.verbose_name)
        assert len(set(names.values())) == len(money_fields)


class TestValidatorOrdering:
    def test_validators_run_in_deterministic_order(self, settings):
        from getpaid.validators import run_getpaid_validators

        call_order = []

        def make_validator(name):
            def validator(data):
                call_order.append(name)
                return data

            return validator

        import sys
        import types

        fake_module = types.ModuleType('fake_validators')
        fake_module.v1 = make_validator('v1')
        fake_module.v2 = make_validator('v2')
        fake_module.v3 = make_validator('v3')
        sys.modules['fake_validators'] = fake_module

        try:
            settings.GETPAID = {
                'VALIDATORS': ['fake_validators.v1', 'fake_validators.v2'],
            }
            settings.GETPAID_BACKEND_SETTINGS = {
                'getpaid.backends.dummy': {'VALIDATORS': ['fake_validators.v3']}
            }

            data = {'backend': 'dummy'}
            for index in range(5):
                call_order.clear()
                run_getpaid_validators(data)
                if index == 0:
                    first_order = call_order.copy()
                else:
                    assert call_order == first_order
        finally:
            del sys.modules['fake_validators']

    def test_validators_no_duplicates(self, settings):
        from getpaid.validators import run_getpaid_validators

        call_count = {'v1': 0}

        def counting_validator(data):
            call_count['v1'] += 1
            return data

        import sys
        import types

        fake_module = types.ModuleType('fake_validators2')
        fake_module.v1 = counting_validator
        sys.modules['fake_validators2'] = fake_module

        try:
            settings.GETPAID = {
                'VALIDATORS': ['fake_validators2.v1'],
            }
            settings.GETPAID_BACKEND_SETTINGS = {
                'getpaid.backends.dummy': {
                    'VALIDATORS': ['fake_validators2.v1']
                }
            }
            run_getpaid_validators({'backend': 'dummy'})
            assert call_count['v1'] == 1
        finally:
            del sys.modules['fake_validators2']


class TestOrderProtocolCompatibility:
    def test_abstract_order_exposes_get_currency(self, order_factory):
        order = order_factory(currency='GBP')

        assert order.get_currency() == 'GBP'



from getpaid.utils import update


class TestUpdate:
    """Tests for the recursive dict update utility."""

    def test_simple_update(self):
        d = {'a': 1, 'b': 2}
        result = update(d, {'b': 3, 'c': 4})
        assert result == {'a': 1, 'b': 3, 'c': 4}
        # Should mutate in-place and return the same dict
        assert result is d

    def test_nested_update(self):
        d = {'a': {'x': 1, 'y': 2}, 'b': 3}
        result = update(d, {'a': {'y': 99, 'z': 100}})
        assert result == {'a': {'x': 1, 'y': 99, 'z': 100}, 'b': 3}

    def test_deep_nested_update(self):
        d = {'a': {'b': {'c': 1}}}
        result = update(d, {'a': {'b': {'d': 2}}})
        assert result == {'a': {'b': {'c': 1, 'd': 2}}}

    def test_overwrite_scalar_with_dict(self):
        d = {'a': 1}
        result = update(d, {'a': {'nested': True}})
        assert result == {'a': {'nested': True}}

    def test_overwrite_dict_with_scalar(self):
        d = {'a': {'nested': True}}
        result = update(d, {'a': 42})
        assert result == {'a': 42}

    def test_empty_update(self):
        d = {'a': 1}
        result = update(d, {})
        assert result == {'a': 1}

    def test_update_empty_dict(self):
        d = {}
        result = update(d, {'a': 1})
        assert result == {'a': 1}

    def test_both_empty(self):
        d = {}
        result = update(d, {})
        assert result == {}



from django import forms
from django.utils.translation import gettext as _


class QuestionForm(forms.Form):
    """
    This dummy form asks for payment authorization.
    """

    authorize_payment = forms.ChoiceField(
        label=_('authorization'), choices=((1, _('yes')), (0, _('no')))
    )

    callback = forms.CharField(widget=forms.HiddenInput, required=False)
    success_url = forms.CharField(widget=forms.HiddenInput, required=False)
    failure_url = forms.CharField(widget=forms.HiddenInput, required=False)



from decimal import Decimal

from django.core.exceptions import ValidationError
from django.forms.models import ModelForm

from .models import Order


class OrderForm(ModelForm):
    class Meta:
        model = Order
        exclude = ('status',)

    def clean_total(self):
        if self.cleaned_data['total'] <= Decimal(0):
            raise ValidationError('Provide some reasonable item price')
        return self.cleaned_data['total']



# Generated by Django 6.0.2 on 2026-02-13 12:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('orders', '0003_fix_amount_refunded_precision'),
    ]

    operations = [
        migrations.AlterField(
            model_name='custompayment',
            name='amount_locked',
            field=models.DecimalField(
                decimal_places=2,
                default=0,
                help_text='Amount locked with this payment, ready to charge.',
                max_digits=20,
                verbose_name='amount locked',
            ),
        ),
        migrations.AlterField(
            model_name='order',
            name='id',
            field=models.AutoField(
                auto_created=True,
                primary_key=True,
                serialize=False,
                verbose_name='ID',
            ),
        ),
    ]



"""Tests for the public package API."""

import tomllib
from pathlib import Path

import getpaid


def test_version() -> None:
    assert getpaid.__version__ == '3.0.0a4'


def test_core_dependency_floor() -> None:
    pyproject_data = tomllib.loads(Path('pyproject.toml').read_text())
    assert (
        'python-getpaid-core>=3.0.0a4'
        in pyproject_data['project']['dependencies']
    )
    assert (
        'python-getpaid-core>=3.0.0a4'
        in pyproject_data['dependency-groups']['dev']
    )



from django.urls import include, path

from . import views
from .registry import registry

app_name = 'getpaid'

urlpatterns = [
    path('new/', views.new_payment, name='create-payment'),
    path(
        'success/<uuid:pk>/',
        views.success,
        name='payment-success',
    ),
    path(
        'failure/<uuid:pk>/',
        views.failure,
        name='payment-failure',
    ),
    path(
        'callback/<uuid:pk>/',
        views.callback,
        name='callback',
    ),
    path('', include(registry.urls)),
]



# Minimalistic settings
import os

from django.urls import reverse_lazy

os.environ['PYTHONBREAKPOINT'] = 'ipdb.set_trace'

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = '=================================================='

DEBUG = True

GETPAID_ORDER_MODEL = 'orders.Order'
GETPAID_PAYMENT_MODEL = 'orders.CustomPayment'

# Configure payment backends via environment variables
# For production: set env vars to override sandbox keys
# For development: sandbox keys are used by default

GETPAID_BACKEND_SETTINGS = {
    'getpaid.backends.dummy': {
        'pos_id': int(os.environ.get('DUMMY_POS_ID', '12345')),
        'second_key': os.environ.get(
            'DUMMY_SECOND_KEY', '91ae651578c5b5aa93f2d38a9be8ce11'
        ),
        'client_id': int(os.environ.get('DUMMY_CLIENT_ID', '12345')),
        'client_secret': os.environ.get(
            'DUMMY_CLIENT_SECRET', '12f071174cb7eb79d4aac5bc2f07563f'
        ),
        'confirmation_method': 'push',
        'gateway': reverse_lazy('paywall:gateway'),
    },
    'getpaid_payu.processor.PayUProcessor': {
        'pos_id': int(os.environ.get('PAYU_POS_ID', '300746')),
        'second_key': os.environ.get(
            'PAYU_SECOND_KEY', 'b6ca15b0d1020e8094d9b5f8d163db54'
        ),
        'oauth_id': int(os.environ.get('PAYU_OAUTH_ID', '300746')),
        'oauth_secret': os.environ.get(
            'PAYU_OAUTH_SECRET', '2ee86a66e5d97e3fadc400c9f19b065d'
        ),
        'sandbox': True,
    },
    'getpaid_paynow.processor.PaynowProcessor': {
        'api_key': os.environ.get(
            'PAYNOW_API_KEY', 'd2e1d881-40b0-4b7e-9168-181bae3dc4e0'
        ),
        'signature_key': os.environ.get(
            'PAYNOW_SIGN_KEY', '8e42a868-5562-440d-817c-4921632fb049'
        ),
        'sandbox': True,
    },
}

PAYWALL_MODE = 'PAY'  # PAY for instant paying, LOCK for pre-auth

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.contenttypes',
    'getpaid',
    'getpaid.backends.dummy',
    'orders',
    'paywall',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

ROOT_URLCONF = 'example.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
            ]
        },
    }
]

WSGI_APPLICATION = 'example.wsgi.application'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

TIME_ZONE = 'UTC'
USE_I18N = False
USE_TZ = True



"""Django-specific plugin registry wrapping getpaid-core."""

import importlib

from django.urls import include, path
from getpaid_core.processor import BaseProcessor
from getpaid_core.registry import PluginRegistry as CorePluginRegistry
from getpaid_core.registry import registry as core_registry


def _importable(name):
    try:
        importlib.import_module(name)
        return True
    except ImportError:
        return False


class DjangoPluginRegistry:
    """Django adapter for getpaid-core's PluginRegistry.

    Adds Django-specific features: URL generation, module-based
    registration (backward compat), and CHOICES-format helpers.
    """

    def __init__(self, core_registry: CorePluginRegistry) -> None:
        self._core = core_registry
        # Map from module path -> slug for backward compat
        self._module_map: dict[str, str] = {}

    def __contains__(self, item):
        return item in self._module_map or self._has_slug(item)

    def __getitem__(self, item):
        return self.get_by_slug(item)

    def get_by_slug(self, item):
        self._sync_aliases_from_core()
        slug = self._module_map.get(item) or item
        return self._core.get_by_slug(slug)

    def resolve_backend(self, item: str) -> str:
        return self.get_by_slug(item).slug

    def get_aliases(self, item: str) -> set[str]:
        processor = self.get_by_slug(item)
        aliases = {processor.slug}
        aliases.update(
            alias
            for alias, slug in self._module_map.items()
            if slug == processor.slug
        )
        return aliases

    def __iter__(self):
        return iter(self._all_keys())

    def register(self, module_or_proc):
        """Register a backend by class or module (backward compat).

        Supports:
        - register(ProcessorClass) -- direct class registration
        - register(module) -- finds module.processor.PaymentProcessor
        """
        if isinstance(module_or_proc, type) and issubclass(
            module_or_proc, BaseProcessor
        ):
            self._core.register(module_or_proc)
            self._remember_aliases(module_or_proc)
            return

        # Module-based registration (v2 backward compat)
        processor = module_or_proc.processor.PaymentProcessor
        self._core.register(processor)
        self._remember_aliases(processor)
        self._module_map[module_or_proc.__name__] = processor.slug

    def unregister(self, slug_or_module_path: str):
        """Remove a backend by slug or module path."""
        slug = self._module_map.pop(slug_or_module_path, slug_or_module_path)
        self._core.unregister(slug)

    def get_choices(self, currency):
        """Get CHOICES for plugins supporting given currency."""
        return self._core.get_choices(currency.upper())

    def get_backends(self, currency):
        """Get backend classes supporting given currency."""
        return self._core.get_for_currency(currency.upper())

    @property
    def urls(self):
        """Provide URL structure for registered plugins with urls modules."""
        result = []
        for module_path, slug in self._module_map.items():
            urls_module = f'{module_path}.urls'
            if _importable(urls_module):
                proc_class = self._core.get_by_slug(slug)
                result.append(
                    path(
                        f'{proc_class.slug}/',
                        include(
                            (urls_module, proc_class.slug),
                            namespace=proc_class.slug,
                        ),
                    )
                )
        return result

    def get_all_supported_currency_choices(self):
        """Get all currencies supported by any plugin, as CHOICES."""
        currencies = self._core.get_all_currencies()
        return [(c.upper(), c.upper()) for c in currencies]

    def _has_slug(self, item):
        try:
            self.get_by_slug(item)
            return True
        except KeyError:
            return False

    def _all_keys(self):
        """Return all keys (both module paths and slugs)."""
        self._sync_aliases_from_core()
        keys = set(self._module_map.keys())
        keys.update(self._core._backends.keys())
        return keys

    def _remember_aliases(self, processor_class):
        class_module = processor_class.__module__
        self._module_map[class_module] = processor_class.slug
        self._module_map[f'{class_module}.{processor_class.__name__}'] = (
            processor_class.slug
        )
        if class_module.endswith('.processor'):
            backend_module = class_module.rsplit('.', 1)[0]
            self._module_map[backend_module] = processor_class.slug

    def _sync_aliases_from_core(self):
        self._core._ensure_discovered()
        for processor_class in self._core._backends.values():
            self._remember_aliases(processor_class)


# Module-level singleton wrapping core's singleton
registry = DjangoPluginRegistry(core_registry)



import json
from decimal import Decimal

import pytest
import swapper
from django.template.response import TemplateResponse

from getpaid.types import PaymentStatus as ps

pytestmark = pytest.mark.django_db

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')


def _make_payment(**kwargs):
    order = Order.objects.create()
    defaults = {
        'order': order,
        'currency': order.currency,
        'amount_required': Decimal(str(order.get_total_amount())),
        'backend': 'getpaid.backends.dummy',
        'description': order.get_description(),
    }
    defaults.update(kwargs)
    return Payment.objects.create(**defaults)


class TestDummyPrepareTransaction:
    def test_rest_method_returns_redirect(self, rf, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'REST'}
        }
        payment = _make_payment()
        result = payment.prepare_transaction(request=rf.get('/'))

        assert result.status_code == 302
        assert payment.status == ps.PREPARED

    def test_get_method_returns_redirect(self, rf, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'GET'}
        }
        payment = _make_payment()
        result = payment.prepare_transaction(request=rf.get('/'))

        assert result.status_code == 302
        assert payment.status == ps.PREPARED

    def test_post_method_returns_template_response(self, rf, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'POST'}
        }
        payment = _make_payment()
        result = payment.prepare_transaction(request=rf.get('/'))

        assert isinstance(result, TemplateResponse)
        assert result.status_code == 200
        assert payment.status == ps.PREPARED


class TestDummyCallbacks:
    def test_paid_callback_marks_payment_paid(self, rf):
        payment = _make_payment(status=ps.PREPARED)
        request = rf.post(
            '',
            data=json.dumps({'new_status': ps.PAID}),
            content_type='application/json',
        )

        response = payment.handle_paywall_callback(request)

        payment.refresh_from_db()
        assert response.status_code == 200
        assert payment.status == ps.PAID
        assert payment.amount_paid == payment.amount_required

    def test_preauth_callback_marks_payment_locked(self, rf):
        payment = _make_payment(status=ps.PREPARED)
        request = rf.post(
            '',
            data=json.dumps({'new_status': ps.PRE_AUTH}),
            content_type='application/json',
        )

        response = payment.handle_paywall_callback(request)

        payment.refresh_from_db()
        assert response.status_code == 200
        assert payment.status == ps.PRE_AUTH
        assert payment.amount_locked == payment.amount_required

    def test_failed_callback_marks_payment_failed(self, rf):
        payment = _make_payment(status=ps.PREPARED)
        request = rf.post(
            '',
            data=json.dumps({'new_status': ps.FAILED}),
            content_type='application/json',
        )

        response = payment.handle_paywall_callback(request)

        payment.refresh_from_db()
        assert response.status_code == 200
        assert payment.status == ps.FAILED

    def test_missing_status_returns_400(self, rf):
        payment = _make_payment(status=ps.PREPARED)
        request = rf.post(
            '', data=json.dumps({}), content_type='application/json'
        )

        response = payment.handle_paywall_callback(request)

        assert response.status_code == 400


class TestDummyPullAndActions:
    def test_fetch_and_update_status_marks_paid(self, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'confirmation_status': 'paid'}
        }
        payment = _make_payment(status=ps.PREPARED)

        result = payment.fetch_and_update_status()

        assert result.status == ps.PAID
        assert result.amount_paid == result.amount_required

    def test_fetch_and_update_status_marks_preauth(self, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'confirmation_status': 'pre_auth'}
        }
        payment = _make_payment(status=ps.PREPARED)

        result = payment.fetch_and_update_status()

        assert result.status == ps.PRE_AUTH
        assert result.amount_locked == result.amount_required

    def test_charge_updates_paid_amount(self):
        payment = _make_payment(
            status=ps.PRE_AUTH,
            amount_locked=Decimal('100.00'),
        )

        result = payment.charge(amount=Decimal('50.00'))

        payment.refresh_from_db()
        assert result.success is True
        assert payment.amount_paid == Decimal('50.00')
        assert payment.amount_locked == Decimal('50.00')
        assert payment.status == ps.PARTIAL

    def test_release_lock_updates_status(self):
        payment = _make_payment(
            status=ps.PRE_AUTH,
            amount_locked=Decimal('100.00'),
        )

        result = payment.release_lock()

        payment.refresh_from_db()
        assert result == Decimal('100.00')
        assert payment.amount_locked == Decimal('0.00')
        assert payment.status == ps.REFUNDED

    def test_start_refund_stores_provider_data(self):
        payment = _make_payment(
            status=ps.PAID,
            amount_paid=Decimal('100.00'),
        )

        result = payment.start_refund(amount=Decimal('50.00'))

        payment.refresh_from_db()
        assert result.amount == Decimal('50.00')
        assert payment.status == ps.REFUND_STARTED
        assert payment.provider_data['refund_id'].startswith('dummy-refund-')

    def test_cancel_refund_restores_paid_state(self):
        payment = _make_payment(
            status=ps.REFUND_STARTED,
            amount_paid=Decimal('100.00'),
            amount_required=Decimal('100.00'),
            provider_data={'refund_id': 'dummy-refund-1'},
        )

        result = payment.cancel_refund()

        payment.refresh_from_db()
        assert result is True
        assert payment.status == ps.PAID



"""Tests that getpaid re-exports core types correctly."""

from decimal import Decimal

from getpaid_core.enums import FraudStatus as CoreFS
from getpaid_core.enums import PaymentStatus as CorePS

from getpaid.exceptions import (
    ChargeFailure,
    CommunicationError,
    GetPaidException,
    InvalidCallbackError,
    InvalidTransitionError,
    LockFailure,
    RefundFailure,
)
from getpaid.status import FraudStatus as StatusFS
from getpaid.status import PaymentStatus as StatusPS
from getpaid.types import (
    BackendMethod,
    BuyerInfo,
    ChargeResponse,
    ConfirmationMethod,
    FraudStatus,
    ItemInfo,
    PaymentStatus,
)


class TestEnumReExports:
    def test_payment_status_values_match_core(self):
        """PaymentStatus values must exactly match core."""
        assert PaymentStatus.NEW == CorePS.NEW
        assert PaymentStatus.PRE_AUTH == CorePS.PRE_AUTH
        assert PaymentStatus.PRE_AUTH == 'pre-auth'
        assert PaymentStatus.IN_CHARGE == CorePS.IN_CHARGE
        assert PaymentStatus.IN_CHARGE == 'charge_started'
        assert PaymentStatus.PARTIAL == CorePS.PARTIAL
        assert PaymentStatus.PARTIAL == 'partially_paid'

    def test_fraud_status_values_match_core(self):
        assert FraudStatus.UNKNOWN == CoreFS.UNKNOWN
        assert FraudStatus.ACCEPTED == CoreFS.ACCEPTED

    def test_payment_status_has_choices(self):
        """Django wrapper must provide .choices and .CHOICES."""
        choices = PaymentStatus.choices
        assert isinstance(choices, tuple)
        assert len(choices) == 9
        # Each choice is (value, label) tuple
        for value, _label in choices:
            assert isinstance(value, str)

    def test_fraud_status_has_choices(self):
        choices = FraudStatus.choices
        assert isinstance(choices, tuple)
        assert len(choices) == 4

    def test_choices_backward_compat(self):
        """CHOICES (uppercase) must equal choices."""
        assert PaymentStatus.choices == PaymentStatus.CHOICES
        assert FraudStatus.choices == FraudStatus.CHOICES

    def test_status_compat_module(self):
        """getpaid.status should re-export the same classes."""
        assert StatusPS is PaymentStatus
        assert StatusFS is FraudStatus

    def test_payment_status_members_all_present(self):
        """All core members must be present in Django enum."""
        for member in CorePS:
            assert hasattr(PaymentStatus, member.name)
            assert PaymentStatus[member.name].value == member.value

    def test_fraud_status_members_all_present(self):
        for member in CoreFS:
            assert hasattr(FraudStatus, member.name)
            assert FraudStatus[member.name].value == member.value


class TestExceptionReExports:
    def test_getpaid_exception_is_core(self):
        """Our GetPaidException should be the core one."""
        from getpaid_core.exceptions import (
            GetPaidException as CoreGPE,
        )

        assert GetPaidException is CoreGPE

    def test_charge_failure_is_communication_error(self):
        assert issubclass(ChargeFailure, CommunicationError)

    def test_lock_failure_is_communication_error(self):
        assert issubclass(LockFailure, CommunicationError)

    def test_refund_failure_is_communication_error(self):
        assert issubclass(RefundFailure, CommunicationError)

    def test_exception_accepts_context_kwarg(self):
        """Backward compat: context kwarg must work."""
        exc = GetPaidException('test', context={'key': 'val'})
        assert exc.context == {'key': 'val'}

    def test_invalid_callback_error(self):
        assert issubclass(InvalidCallbackError, GetPaidException)

    def test_invalid_transition_error(self):
        assert issubclass(InvalidTransitionError, GetPaidException)


class TestTypeReExports:
    def test_buyer_info_is_typed_dict(self):
        info: BuyerInfo = {'email': 'a@b.com'}
        assert 'email' in info

    def test_item_info(self):
        item: ItemInfo = {
            'name': 'x',
            'quantity': 1,
            'unit_price': Decimal(10),
        }
        assert item['name'] == 'x'

    def test_charge_response(self):
        resp = ChargeResponse(
            amount_charged=Decimal(10),
            success=True,
            async_call=False,
        )
        assert resp.success is True

    def test_backend_method_enum(self):
        assert BackendMethod.GET == 'GET'
        assert BackendMethod.POST == 'POST'
        assert BackendMethod.REST == 'REST'

    def test_confirmation_method_enum(self):
        assert ConfirmationMethod.PUSH == 'PUSH'
        assert ConfirmationMethod.PULL == 'PULL'



from pathlib import Path

from tests.database import get_test_databases

BASE_DIR = Path(__file__).resolve().parent.parent
SECRET_KEY = 'tests-default-payment-secret-key'
DEBUG = True
USE_I18N = False
USE_TZ = True
TIME_ZONE = 'UTC'
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

GETPAID_ORDER_MODEL = 'default_order_app.Order'
GETPAID_PAYMENT_MODEL = 'getpaid.Payment'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'getpaid',
    'tests.default_order_app',
]

MIDDLEWARE = []
ROOT_URLCONF = 'tests.urls_default_payment'
TEMPLATES = []

DATABASES = get_test_databases()



import pytest
import swapper

from getpaid.status import FraudStatus as fs
from getpaid.status import PaymentStatus as ps

pytestmark = pytest.mark.django_db

Payment = swapper.load_model('getpaid', 'Payment')


def test_payment_status_values_available(payment_factory):
    payment = payment_factory()
    assert payment.status == ps.NEW
    assert ps.PREPARED == 'prepared'
    assert ps.PRE_AUTH == 'pre-auth'
    assert ps.PAID == 'paid'


def test_flag_as_fraud_sets_rejected(payment_factory):
    payment = payment_factory()

    payment.flag_as_fraud(message='Suspicious IP')
    payment.save()

    assert payment.fraud_status == fs.REJECTED
    assert payment.fraud_message == 'Suspicious IP'


def test_flag_as_legit_sets_accepted(payment_factory):
    payment = payment_factory()

    payment.flag_as_legit(message='Verified buyer')
    payment.save()

    assert payment.fraud_status == fs.ACCEPTED
    assert payment.fraud_message == 'Verified buyer'


def test_flag_for_check_sets_check(payment_factory):
    payment = payment_factory()

    payment.flag_for_check(message='Needs review')
    payment.save()

    assert payment.fraud_status == fs.CHECK
    assert payment.fraud_message == 'Needs review'



import uuid

import pytest
import swapper
from django.template.response import TemplateResponse

from getpaid.registry import registry
from getpaid.types import PaymentStatus as ps

from .tools import Plugin

pytestmark = pytest.mark.django_db
dummy = 'getpaid.backends.dummy'

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')


def _prep_conf(paywall_method='REST', confirmation_status='paid'):
    return {
        'getpaid.backends.dummy': {
            'paywall_method': paywall_method,
            'confirmation_status': confirmation_status,
        }
    }


class TestModelProcessor:
    @pytest.fixture(autouse=True)
    def setup_plugin(self):
        if Plugin.slug not in registry:
            registry.register(Plugin)

    def test_model_and_dummy_backend(self):
        order = Order.objects.create()
        payment = Payment.objects.create(
            order=order,
            currency=order.currency,
            amount_required=order.get_total_amount(),
            backend=dummy,
            description=order.get_description(),
        )
        proc = payment._get_processor()
        assert isinstance(proc, registry[dummy])

    def test_model_and_test_backend(self):
        order = Order.objects.create()
        payment = Payment.objects.create(
            order=order,
            currency=order.currency,
            amount_required=order.get_total_amount(),
            backend=Plugin.slug,
            description=order.get_description(),
        )
        proc = payment._get_processor()
        assert isinstance(proc, registry[Plugin.slug])


def test_get_flow_begin(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(paywall_method='GET')
    payment = payment_factory(external_id=uuid.uuid4())

    result = payment.prepare_transaction(None)
    assert result.status_code == 302
    payment.refresh_from_db()
    assert payment.status == ps.PREPARED


def test_post_flow_begin(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(paywall_method='POST')
    payment = payment_factory(external_id=uuid.uuid4())

    result = payment.prepare_transaction(rf.get('/'))
    assert isinstance(result, TemplateResponse)
    payment.refresh_from_db()
    assert payment.status == ps.PREPARED


def test_rest_flow_begin(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(paywall_method='REST')
    payment = payment_factory(external_id=uuid.uuid4())

    result = payment.prepare_transaction(None)
    assert result.status_code == 302
    payment.refresh_from_db()
    assert payment.status == ps.PREPARED


def test_pull_flow_paid(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(confirmation_status='paid')
    payment = payment_factory(status=ps.PREPARED, external_id=uuid.uuid4())

    payment.fetch_and_update_status()
    payment.refresh_from_db()

    assert payment.status == ps.PAID


def test_pull_flow_locked(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(
        confirmation_status='pre_auth'
    )
    payment = payment_factory(status=ps.PREPARED, external_id=uuid.uuid4())

    payment.fetch_and_update_status()
    payment.refresh_from_db()

    assert payment.status == ps.PRE_AUTH


def test_pull_flow_failed(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(confirmation_status='failed')
    payment = payment_factory(status=ps.PREPARED, external_id=uuid.uuid4())

    payment.fetch_and_update_status()
    payment.refresh_from_db()

    assert payment.status == ps.FAILED



import json
from decimal import Decimal
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
import swapper
from django.http import HttpResponse, HttpResponseForbidden
from django.test import RequestFactory

from getpaid.exceptions import GetPaidException
from getpaid.processor import BaseProcessor
from getpaid.views import CallbackDetailView


class ConcreteProcessor(BaseProcessor):
    """Minimal concrete processor for testing."""

    display_name = 'Test'
    accepted_currencies = ['EUR']
    slug = 'test_proc'

    def prepare_transaction(self, request=None, view=None, **kwargs):
        return MagicMock(status_code=302, url='/redirect')


class VerifyingProcessor(ConcreteProcessor):
    """Processor that implements verify_callback."""

    slug = 'verifying_proc'

    def verify_callback(self, request, **kwargs):
        sig = request.headers.get('X-Signature', '')
        if sig != 'valid-signature':
            raise GetPaidException('Invalid callback signature')


class TestVerifyCallback:
    def test_base_processor_has_verify_callback(self):
        assert hasattr(BaseProcessor, 'verify_callback')

    def test_base_verify_callback_is_noop_by_default(self):
        mock_payment = MagicMock()
        mock_payment.backend = 'test_proc'
        proc = ConcreteProcessor(mock_payment)
        mock_request = MagicMock()
        proc.verify_callback(mock_request)

    def test_custom_verify_callback_rejects_invalid(self):
        mock_payment = MagicMock()
        mock_payment.backend = 'verifying_proc'
        proc = VerifyingProcessor(mock_payment)

        bad_request = MagicMock()
        bad_request.headers = {'X-Signature': 'bad'}

        with pytest.raises(
            GetPaidException, match='Invalid callback signature'
        ):
            proc.verify_callback(bad_request)

    def test_custom_verify_callback_accepts_valid(self):
        mock_payment = MagicMock()
        mock_payment.backend = 'verifying_proc'
        proc = VerifyingProcessor(mock_payment)

        good_request = MagicMock()
        good_request.headers = {'X-Signature': 'valid-signature'}

        proc.verify_callback(good_request)


@pytest.mark.django_db
class TestGetOurBaseurl:
    def test_with_request_uses_request_host(self, rf, settings):
        settings.DEBUG = True
        request = rf.get('/')
        url = BaseProcessor.get_our_baseurl(request)
        assert url.startswith('http')
        assert '://' in url
        assert url.endswith('/')

    def test_without_request_uses_sites_framework(self, settings):
        from django.contrib.sites.models import Site

        Site.objects.update_or_create(
            id=settings.SITE_ID,
            defaults={'domain': 'example.com', 'name': 'Example'},
        )
        settings.DEBUG = False
        url = BaseProcessor.get_our_baseurl(request=None)
        assert 'example.com' in url
        assert url.startswith('https://')
        assert url.endswith('/')

    def test_without_request_no_hardcoded_localhost(self, settings):
        from django.contrib.sites.models import Site

        Site.objects.update_or_create(
            id=settings.SITE_ID,
            defaults={'domain': 'myshop.com', 'name': 'My Shop'},
        )
        url = BaseProcessor.get_our_baseurl(request=None)
        assert '127.0.0.1' not in url

    def test_production_uses_https(self, rf, settings):
        settings.DEBUG = False
        request = rf.get('/')
        url = BaseProcessor.get_our_baseurl(request)
        assert url.startswith('https://')

    def test_debug_uses_http(self, rf, settings):
        settings.DEBUG = True
        request = rf.get('/')
        url = BaseProcessor.get_our_baseurl(request)
        assert url.startswith('http://')


@pytest.mark.django_db
class TestCallbackViewSecurity:
    @pytest.fixture(autouse=True)
    def setup_payment(self):
        from tests.factories import OrderFactory

        self.factory = RequestFactory()
        self.order = OrderFactory()
        Payment = swapper.load_model('getpaid', 'Payment')
        self.payment = Payment.objects.create(
            order=self.order,
            amount_required=Decimal(str(self.order.get_total_amount())),
            currency=self.order.currency,
            description='Test payment',
            backend='getpaid.backends.dummy',
        )

    def test_callback_view_calls_verify_callback(self):
        request = self.factory.post(f'/payments/callback/{self.payment.pk}/')

        mock_processor = MagicMock()
        mock_processor.verify_callback = MagicMock()
        mock_processor.handle_paywall_callback = MagicMock(
            return_value=HttpResponse(b'OK')
        )
        del mock_processor.handle_callback

        with patch.object(
            type(self.payment),
            '_get_processor',
            return_value=mock_processor,
        ):
            view = CallbackDetailView()
            view.request = request
            response = view.post(request, pk=self.payment.pk)

        mock_processor.verify_callback.assert_called_once_with(request)
        assert response.status_code == 200

    def test_callback_view_returns_403_on_verify_failure(self):
        request = self.factory.post(f'/payments/callback/{self.payment.pk}/')

        mock_processor = MagicMock()
        mock_processor.verify_callback = MagicMock(
            side_effect=GetPaidException('Bad signature')
        )
        mock_processor.handle_paywall_callback = MagicMock(
            return_value=HttpResponse(b'OK')
        )
        del mock_processor.handle_callback

        with patch.object(
            type(self.payment),
            '_get_processor',
            return_value=mock_processor,
        ):
            view = CallbackDetailView()
            view.request = request
            response = view.post(request, pk=self.payment.pk)

        assert response.status_code == 403
        mock_processor.handle_paywall_callback.assert_not_called()

    def test_callback_view_proceeds_when_verify_passes(self):
        request = self.factory.post(f'/payments/callback/{self.payment.pk}/')

        mock_processor = MagicMock()
        mock_processor.verify_callback = MagicMock()
        mock_processor.handle_paywall_callback = MagicMock(
            return_value=HttpResponse(b'OK', status=200)
        )
        del mock_processor.handle_callback

        with patch.object(
            type(self.payment),
            '_get_processor',
            return_value=mock_processor,
        ):
            view = CallbackDetailView()
            view.request = request
            response = view.post(request, pk=self.payment.pk)

        mock_processor.verify_callback.assert_called_once_with(request)
        mock_processor.handle_paywall_callback.assert_called_once()
        assert response.status_code == 200

    def test_callback_view_adapts_async_processor_contract(self):
        payload = {'status': 'COMPLETED', 'orderId': 'PAYU-1'}
        request = self.factory.post(
            f'/payments/callback/{self.payment.pk}/',
            data=json.dumps(payload),
            content_type='application/json',
            HTTP_X_SIGNATURE='valid-signature',
        )

        mock_processor = MagicMock()
        mock_processor.verify_callback = AsyncMock()
        mock_processor.handle_callback = AsyncMock(return_value=None)

        with patch.object(
            type(self.payment),
            '_get_processor',
            return_value=mock_processor,
        ):
            view = CallbackDetailView()
            view.request = request
            response = view.post(request, pk=self.payment.pk)

        verify_args = mock_processor.verify_callback.await_args
        handle_args = mock_processor.handle_callback.await_args
        assert verify_args.args[0] == payload
        assert verify_args.args[1]['X-SIGNATURE'] == 'valid-signature'
        assert verify_args.kwargs['raw_body'] == request.body
        assert handle_args.args[0] == payload
        assert response.status_code == 200

    def test_callback_view_returns_403_for_async_verify_failure(self):
        request = self.factory.post(
            f'/payments/callback/{self.payment.pk}/',
            data=json.dumps({'status': 'COMPLETED'}),
            content_type='application/json',
            HTTP_X_SIGNATURE='bad-signature',
        )

        mock_processor = MagicMock()
        mock_processor.verify_callback = AsyncMock(
            side_effect=GetPaidException('Bad signature')
        )
        mock_processor.handle_callback = AsyncMock(return_value=None)

        with patch.object(
            type(self.payment),
            '_get_processor',
            return_value=mock_processor,
        ):
            view = CallbackDetailView()
            view.request = request
            response = view.post(request, pk=self.payment.pk)

        assert isinstance(response, HttpResponseForbidden)
        mock_processor.handle_callback.assert_not_called()



"""Tests for callback adapter."""

import json
from unittest.mock import AsyncMock, Mock

import pytest
from django.test import RequestFactory

from getpaid.adapters import (
    adapt_callback_request,
    call_processor_verify_callback,
)


@pytest.fixture
def factory():
    return RequestFactory()


def test_adapt_callback_request_json(factory):
    """Test JSON payload extraction."""
    payload = {'status': 'completed', 'transaction_id': '12345'}
    request = factory.post(
        '/callback/',
        data=json.dumps(payload),
        content_type='application/json',
        HTTP_X_SIGNATURE='abc123',
        HTTP_AUTHORIZATION='Bearer token',
    )

    data, headers, raw_body = adapt_callback_request(request)

    assert data == payload
    assert headers['X-SIGNATURE'] == 'abc123'
    assert headers['AUTHORIZATION'] == 'Bearer token'
    assert raw_body == request.body


def test_adapt_callback_request_form_data(factory):
    """Test form data extraction."""
    request = factory.post(
        '/callback/',
        data={'status': 'paid', 'amount': '100'},
        HTTP_X_CUSTOM='value',
    )

    data, headers, _raw_body = adapt_callback_request(request)

    assert 'status' in data
    assert headers['X-CUSTOM'] == 'value'


def test_adapt_callback_request_flattens_single_form_values(factory):
    request = factory.post(
        '/callback/',
        data={'status': 'paid', 'amount': '100'},
    )

    data, _headers, _raw_body = adapt_callback_request(request)

    assert data == {'status': 'paid', 'amount': '100'}


def test_call_processor_verify_callback_async():
    """Test async processor verification."""
    processor = Mock()
    processor.verify_callback = AsyncMock()

    factory = RequestFactory()
    request = factory.post(
        '/callback/',
        data=json.dumps({'test': 'data'}),
        content_type='application/json',
    )

    call_processor_verify_callback(processor, request)

    # Should have been called with (data, headers, raw_body=...)
    processor.verify_callback.assert_called_once()
    call_args = processor.verify_callback.call_args
    assert call_args[0][0] == {'test': 'data'}  # data
    assert isinstance(call_args[0][1], dict)  # headers
    assert 'raw_body' in call_args[1]  # kwargs


def test_call_processor_verify_callback_sync():
    """Test sync processor verification (backward compat)."""
    processor = Mock()
    processor.verify_callback = Mock()  # Sync method

    factory = RequestFactory()
    request = factory.post('/callback/', data={'test': 'data'})

    call_processor_verify_callback(processor, request)

    # Should have been called with HttpRequest directly
    processor.verify_callback.assert_called_once_with(request)



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from getpaid import PaymentStatus

logger = logging.getLogger('getpaid_example')


@receiver(post_save)
def payment_status_changed_listener(sender, instance, **kwargs):
    """Update order status when payment is completed."""
    # Only act on Payment models
    if not hasattr(instance, 'order') or not hasattr(instance, 'status'):
        return

    from getpaid.abstracts import AbstractPayment

    if not isinstance(instance, AbstractPayment):
        return

    if instance.status == PaymentStatus.PAID:
        logger.debug(
            'Payment %s is PAID, updating order status.',
            instance.id,
        )
        instance.order.status = 'P'
        instance.order.save()



from collections.abc import Mapping


def update(d: dict, u: dict) -> dict:
    """
    Handy tool to recursively update dicts.
    """
    for k, v in u.items():
        existing = d.get(k)
        if isinstance(v, Mapping) and isinstance(existing, dict):
            update(existing, v)
        else:
            d[k] = v
    return d



# Generated by Django 3.0.5 on 2020-04-12 19:28

import uuid

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Order',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        default='Lock, Stock and Two Smoking Barrels',
                        max_length=100,
                    ),
                ),
                (
                    'total',
                    models.DecimalField(
                        decimal_places=2, default='199.99', max_digits=8
                    ),
                ),
                ('currency', models.CharField(default='EUR', max_length=3)),
                (
                    'status',
                    models.CharField(
                        blank=True,
                        choices=[
                            ('W', 'Waiting for payment'),
                            ('P', 'Payment complete'),
                        ],
                        default='W',
                        max_length=1,
                    ),
                ),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CustomPayment',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'amount_required',
                    models.DecimalField(
                        decimal_places=2,
                        help_text='Amount required to fulfill the payment; in selected currency, normal notation',
                        max_digits=20,
                        verbose_name='amount required',
                    ),
                ),
                (
                    'currency',
                    models.CharField(max_length=3, verbose_name='currency'),
                ),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('new', 'new'),
                            ('in_progress', 'in progress'),
                            ('accepted_for_proc', 'accepted for processing'),
                            ('partially_paid', 'partially paid'),
                            ('paid', 'paid'),
                            ('cancelled', 'cancelled'),
                            ('failed', 'failed'),
                            ('refund_started', 'refund started'),
                            ('refunded', 'refunded'),
                        ],
                        db_index=True,
                        default='new',
                        max_length=20,
                        verbose_name='status',
                    ),
                ),
                (
                    'backend',
                    models.CharField(
                        db_index=True, max_length=100, verbose_name='backend'
                    ),
                ),
                (
                    'created_on',
                    models.DateTimeField(
                        auto_now_add=True,
                        db_index=True,
                        verbose_name='created on',
                    ),
                ),
                (
                    'last_payment_on',
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=None,
                        null=True,
                        verbose_name='paid on',
                    ),
                ),
                (
                    'amount_locked',
                    models.DecimalField(
                        decimal_places=2,
                        default=0,
                        help_text='Amount locked with this payment, ready to charge.',
                        max_digits=20,
                        verbose_name='amount paid',
                    ),
                ),
                (
                    'amount_paid',
                    models.DecimalField(
                        decimal_places=2,
                        default=0,
                        help_text='Amount actually paid.',
                        max_digits=20,
                        verbose_name='amount paid',
                    ),
                ),
                (
                    'refunded_on',
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=None,
                        null=True,
                        verbose_name='refunded on',
                    ),
                ),
                (
                    'amount_refunded',
                    models.DecimalField(
                        decimal_places=4,
                        default=0,
                        max_digits=20,
                        verbose_name='amount refunded',
                    ),
                ),
                (
                    'external_id',
                    models.CharField(
                        blank=True,
                        db_index=True,
                        default='',
                        max_length=64,
                        verbose_name='external id',
                    ),
                ),
                (
                    'description',
                    models.CharField(
                        blank=True,
                        default='',
                        max_length=128,
                        verbose_name='description',
                    ),
                ),
                (
                    'fraud_status',
                    models.CharField(
                        choices=[
                            ('unknown', 'unknown'),
                            ('accepted', 'accepted'),
                            ('rejected', 'rejected'),
                            ('check', 'check'),
                        ],
                        db_index=True,
                        default='unknown',
                        max_length=20,
                        verbose_name='fraud status',
                    ),
                ),
                (
                    'fraud_message',
                    models.TextField(blank=True, verbose_name='fraud message'),
                ),
                ('custom', models.BooleanField(default=True, editable=False)),
                (
                    'order',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='payments',
                        to='orders.Order',
                        verbose_name='order',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Payment',
                'verbose_name_plural': 'Payments',
                'ordering': ['-created_on'],
                'abstract': False,
            },
        ),
    ]



# Create your views here.
from django import http
from django.views import View
from django.views.generic import CreateView
from django.views.generic.detail import DetailView

from getpaid.forms import PaymentMethodForm

from .forms import OrderForm
from .models import Order


class HomeView(CreateView):
    model = Order
    template_name = 'home.html'
    form_class = OrderForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['orders'] = Order.objects.all()
        return context


class OrderView(DetailView):
    model = Order

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['payment_form'] = PaymentMethodForm(
            initial={'order': self.object, 'currency': self.object.currency}
        )
        return context


class PostGetter(View):
    def post(self, request, *args, **kwargs):
        return http.HttpResponse('OK')



import os
import subprocess  # noqa: S404
import sys
from pathlib import Path


def test_default_payment_model_has_no_pending_migrations():
    project_root = Path(__file__).resolve().parents[1]
    env = os.environ.copy()
    env['DJANGO_SETTINGS_MODULE'] = 'tests.settings_default_payment'
    env['PYTHONPATH'] = os.pathsep.join([
        str(project_root),
        *filter(None, [env.get('PYTHONPATH')]),
    ])

    result = subprocess.run(  # noqa: S603
        [
            sys.executable,
            'manage.py',
            'makemigrations',
            '--check',
            '--dry-run',
            'getpaid',
        ],
        cwd=project_root,
        capture_output=True,
        text=True,
        check=False,
        env=env,
    )

    output = '\n'.join(
        part for part in (result.stdout, result.stderr) if part.strip()
    )
    assert result.returncode == 0, output



import uuid

import httpx
from django.db import models

from getpaid.status import FraudStatus as fs
from getpaid.status import PaymentStatus as ps


class PaymentEntry(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    ext_id = models.CharField(max_length=100, db_index=True, default=uuid.uuid4)
    value = models.DecimalField(decimal_places=2, max_digits=20)
    currency = models.CharField(max_length=3)
    description = models.TextField(blank=True)
    callback = models.URLField(blank=True)
    success_url = models.URLField(blank=True)
    failure_url = models.URLField(blank=True)
    payment_status = models.CharField(
        max_length=50,
        choices=ps.choices,
        default=ps.PREPARED,
    )
    fraud_status = models.CharField(
        max_length=50,
        choices=fs.choices,
        default=fs.UNKNOWN,
    )

    def _send_status_to_callback(self, status):
        httpx.post(
            self.callback, json={'id': str(self.id), 'new_status': status}
        )

    def send_confirm_lock(self):
        self.payment_status = ps.PRE_AUTH
        self.save()
        self._send_status_to_callback(ps.PRE_AUTH)

    def send_confirm_charge(self):
        self.payment_status = ps.PAID
        self.save()
        self._send_status_to_callback(ps.PAID)

    def send_fail(self):
        self.payment_status = ps.FAILED
        self.save()
        self._send_status_to_callback(ps.FAILED)

    def start_refund(self):
        self.payment_status = ps.REFUND_STARTED
        self.save()

    def send_confirm_refund(self):
        self.payment_status = ps.REFUNDED
        self.save()
        self._send_status_to_callback(ps.REFUNDED)

    def cancel_refund(self):
        self.payment_status = ps.PAID
        self.save()



# Minimalistic settings
import os

from tests.database import get_test_databases

os.environ['PYTHONBREAKPOINT'] = 'ipdb.set_trace'

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = '=================================================='

DEBUG = True

GETPAID_ORDER_MODEL = 'orders.Order'
GETPAID_PAYMENT_MODEL = 'orders.CustomPayment'
CONFIRMATION_METHOD = 'PUSH'
REGISTRATION_METHOD = 'REST'

GETPAID_BACKEND_SETTINGS = {
    'getpaid.backends.dummy': {
        'confirmation_method': CONFIRMATION_METHOD,
        'paywall_method': REGISTRATION_METHOD,
        # "push" for automatic callback,
        # "pull" if you want to call fetch_status separately
    },
}

PAYWALL_MODE = 'LOCK'  # PAY for instant paying, LOCK for pre-auth

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.contenttypes',
    'django.contrib.sites',
    'getpaid',
    'getpaid.backends.dummy',
    'orders',
    'paywall',
]

SITE_ID = 1

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

ROOT_URLCONF = 'example.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
            ]
        },
    }
]

WSGI_APPLICATION = 'example.wsgi.application'

DATABASES = get_test_databases()

TIME_ZONE = 'UTC'
USE_I18N = False
USE_TZ = True



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _


class DummyPluginAppConfig(AppConfig):
    name = 'getpaid.backends.dummy'
    label = 'getpaid_dummy'
    verbose_name = _('Dummy paywall')
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        if not settings.DEBUG:
            raise ImproperlyConfigured('Do not use dummy plugin on production!')

        from getpaid.backends.dummy.processor import PaymentProcessor
        from getpaid.registry import registry

        registry.register(PaymentProcessor)



from django.contrib import admin

from . import models


# Payment model is used here directly so that this PaymentAdmin does not show
# if Payment is swapped.
@admin.register(models.Payment)
class PaymentAdmin(admin.ModelAdmin):
    list_display = (
        'id',
        'order_id',
        'amount_required',
        'currency',
        'status',
        'backend',
        'created_on',
        'last_payment_on',
        'amount_paid',
    )
    search_fields = ('id', 'order_id')
    date_hierarchy = 'created_on'



# Generated by Django 3.0.5 on 2020-04-12 19:28

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='PaymentEntry',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ('payment', models.CharField(max_length=100)),
                ('value', models.DecimalField(decimal_places=2, max_digits=20)),
                ('currency', models.CharField(max_length=3)),
                ('description', models.TextField()),
                ('callback', models.URLField()),
                ('success_url', models.URLField()),
                ('failure_url', models.URLField()),
            ],
        ),
    ]



"""Adapters to bridge Django sync views to core async processors."""

import inspect
import json
from typing import Any

from asgiref.sync import async_to_sync
from django.http import HttpRequest


def adapt_callback_request(
    request: HttpRequest,
) -> tuple[dict[str, Any], dict[str, str], bytes]:
    """Extract (data, headers, raw_body) from Django HttpRequest.

    Returns:
        (data, headers, raw_body) tuple suitable for core processor.verify_callback()
    """
    # Capture raw_body first before accessing POST (which consumes the stream)
    raw_body = request.body

    # Extract data from request body
    if request.content_type and 'json' in request.content_type:
        data = json.loads(raw_body)
    else:
        data = {}
        for key in request.POST:
            values = list(request.POST.getlist(key))
            data[key] = values[0] if len(values) == 1 else values

    # Extract headers (HTTP_X_FOO → X-Foo)
    headers = {
        key.replace('HTTP_', '', 1).replace('_', '-'): value
        for key, value in request.META.items()
        if key.startswith('HTTP_')
    }

    return data, headers, raw_body


def call_processor_verify_callback(
    processor: Any, request: HttpRequest
) -> None:
    """Call processor.verify_callback, handling async/sync bridge.

    Detects if processor has async verify_callback and uses async_to_sync if needed.
    Falls back to Django-style verify_callback(request) for backward compat.
    """
    verify_method = getattr(processor, 'verify_callback', None)

    if verify_method is None:
        return  # No verification needed

    # Check if it's an async method
    if inspect.iscoroutinefunction(verify_method):
        # Core-style: async def verify_callback(data, headers, **kwargs)
        data, headers, raw_body = adapt_callback_request(request)
        async_to_sync(verify_method)(data, headers, raw_body=raw_body)
    else:
        # Django-style (backward compat): def verify_callback(request)
        verify_method(request)



# Generated by Django 3.0.5 on 2020-04-19 15:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('paywall', '0002_auto_20200417_2107'),
    ]

    operations = [
        migrations.AlterField(
            model_name='paymententry',
            name='callback',
            field=models.URLField(blank=True),
        ),
        migrations.AlterField(
            model_name='paymententry',
            name='description',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='paymententry',
            name='failure_url',
            field=models.URLField(blank=True),
        ),
        migrations.AlterField(
            model_name='paymententry',
            name='success_url',
            field=models.URLField(blank=True),
        ),
    ]



from decimal import Decimal

import pytest
from django.test import override_settings

from getpaid.forms import PaymentMethodForm

pytestmark = pytest.mark.django_db


@override_settings(
    GETPAID_BACKEND_SETTINGS={
        'getpaid.backends.dummy': {'paywall_method': 'REST'}
    }
)
def test_payment_method_form_recomputes_server_controlled_fields(order_factory):
    order = order_factory(total=Decimal('10.00'), currency='EUR')

    form = PaymentMethodForm(
        data={
            'order': order.pk,
            'amount_required': '999.99',
            'description': 'Tampered description',
            'currency': 'USD',
            'backend': 'dummy',
        },
        initial={'order': order, 'currency': order.currency},
    )

    assert form.is_valid(), form.errors

    assert form.cleaned_data['amount_required'] == Decimal('10.00')
    assert form.cleaned_data['description'] == order.get_description()
    assert form.cleaned_data['currency'] == 'EUR'


@override_settings(
    GETPAID_BACKEND_SETTINGS={
        'getpaid.backends.dummy': {'paywall_method': 'REST'}
    }
)
def test_payment_method_form_save_persists_server_controlled_fields(
    order_factory,
):
    order = order_factory(total=Decimal('25.50'), currency='USD')

    form = PaymentMethodForm(
        data={
            'order': order.pk,
            'amount_required': '1.00',
            'description': 'Tampered description',
            'currency': 'EUR',
            'backend': 'dummy',
        },
        initial={'order': order, 'currency': order.currency},
    )

    assert form.is_valid(), form.errors

    payment = form.save()

    assert payment.amount_required == Decimal('25.50')
    assert payment.description == order.get_description()
    assert payment.currency == 'USD'
    assert payment.backend == 'dummy'



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('getpaid', '0003_remove_fsm_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='payment',
            name='provider_data',
            field=models.JSONField(
                blank=True,
                default=dict,
                verbose_name='provider data',
            ),
        ),
    ]



from django.urls import include, path
from orders.views import OrderView

urlpatterns = [
    path('order/<int:pk>/', OrderView.as_view(), name='order_detail'),
    path('payments/', include('getpaid.urls')),
    path('paywall/', include('paywall.urls')),
]



import json

import swapper
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404
from django.views import View


class CallbackView(View):
    """
    Callback endpoint for the dummy backend's PUSH confirmation flow.

    Accepts POST requests with a JSON body containing:
        - payment_id: the payment's external_id (UUID)
        - new_status: the new payment status to apply

    Delegates actual processing to the payment's handle_paywall_callback.
    """

    def post(self, request, *args, **kwargs):
        try:
            data = json.loads(request.body)
        except (json.JSONDecodeError, ValueError):
            return HttpResponseBadRequest('Invalid JSON')

        payment_id = data.get('payment_id')
        if not payment_id:
            return HttpResponseBadRequest('Missing payment_id')

        Payment = swapper.load_model('getpaid', 'Payment')
        payment = get_object_or_404(
            Payment,
            external_id=payment_id,
            backend='getpaid.backends.dummy',
        )
        return payment.handle_paywall_callback(request, *args, **kwargs)



"""Payment views for django-getpaid v3."""

import logging

import swapper
from django import http
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import CreateView, RedirectView

from .exceptions import GetPaidException
from .forms import PaymentMethodForm
from .runtime import handle_callback_request

logger = logging.getLogger(__name__)


class CreatePaymentView(LoginRequiredMixin, CreateView):
    model = swapper.load_model('getpaid', 'Payment')
    form_class = PaymentMethodForm

    def get(self, request, *args, **kwargs):
        return http.HttpResponseNotAllowed(['POST'])

    def form_valid(self, form):
        payment = form.save()
        return payment.prepare_transaction(request=self.request, view=self)

    def form_invalid(self, form):
        return super().form_invalid(form)


new_payment = CreatePaymentView.as_view()


class FallbackView(RedirectView):
    """Return view from paywall after payment completion/rejection."""

    success = None
    permanent = False

    def get_redirect_url(self, *args, **kwargs):
        Payment = swapper.load_model('getpaid', 'Payment')
        payment = get_object_or_404(Payment, pk=self.kwargs['pk'])
        return payment.get_return_redirect_url(
            request=self.request, success=self.success
        )


class SuccessView(FallbackView):
    success = True


success = SuccessView.as_view()


class FailureView(FallbackView):
    success = False


failure = FailureView.as_view()


class CallbackDetailView(View):
    """Handle paywall callback via PUSH flow."""

    def post(self, request: HttpRequest, pk, *args, **kwargs):
        Payment = swapper.load_model('getpaid', 'Payment')
        payment = get_object_or_404(Payment, pk=pk)
        try:
            return handle_callback_request(payment, request, **kwargs)
        except GetPaidException:
            logger.warning('Callback verification failed for payment %s', pk)
            return http.HttpResponseForbidden(b'Callback verification failed')


callback = csrf_exempt(CallbackDetailView.as_view())



"""Tests for test database configuration helpers."""

import pytest


def test_database_settings_default_to_sqlite(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from tests.database import get_test_databases

    monkeypatch.delenv('TEST_DATABASE_URL', raising=False)

    assert get_test_databases() == {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': ':memory:',
        }
    }


def test_database_settings_use_postgres_environment_override(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from tests.database import get_test_databases

    monkeypatch.setenv(
        'TEST_DATABASE_URL',
        'postgres://test_user:test_password@testdb:5432/test_db',
    )

    assert get_test_databases() == {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': 'test_db',
            'USER': 'test_user',
            'PASSWORD': 'test_password',
            'HOST': 'testdb',
            'PORT': 5432,
        }
    }



from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Order',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'description',
                    models.CharField(default='Test order', max_length=100),
                ),
                (
                    'total',
                    models.DecimalField(
                        decimal_places=2,
                        default='10.00',
                        max_digits=8,
                    ),
                ),
                ('currency', models.CharField(default='EUR', max_length=3)),
            ],
            options={'abstract': False},
        ),
    ]



"""Shared test database settings helpers."""

from __future__ import annotations

import os
from urllib.parse import urlparse


def get_test_databases() -> dict[str, dict[str, object]]:
    database_url = os.environ.get('TEST_DATABASE_URL')
    if not database_url:
        return {
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',
            }
        }

    parsed = urlparse(database_url)
    engine = _django_engine(parsed.scheme)
    return {
        'default': {
            'ENGINE': engine,
            'NAME': parsed.path.lstrip('/'),
            'USER': parsed.username or '',
            'PASSWORD': parsed.password or '',
            'HOST': parsed.hostname or '',
            'PORT': parsed.port or '',
        }
    }


def _django_engine(scheme: str) -> str:
    normalized = scheme.lower()
    if normalized in {'postgres', 'postgresql'}:
        return 'django.db.backends.postgresql'
    raise ValueError(f'Unsupported TEST_DATABASE_URL scheme: {scheme}')



from django.apps import AppConfig


class Config(AppConfig):
    name = 'orders'
    verbose_name = 'application orders'
    label = 'orders'
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        # noinspection PyUnresolvedReferences
        from . import signals  # noqa: F401



"""Tests for deprecated API cleanup (Phase 5).

Ensures that:
- No `default_app_config` in __init__.py files
- No `typing_extensions` dependency for TypedDict
- No try/except fallback for classproperty
- AppConfig classes set `default_auto_field`
"""

import importlib
import inspect
import pathlib


class TestNoDefaultAppConfig:
    """default_app_config is deprecated since Django 3.2 and no-op in 6.0."""

    def test_getpaid_init_no_default_app_config(self):
        import getpaid

        assert not hasattr(getpaid, 'default_app_config') or (
            'default_app_config' not in inspect.getsource(getpaid)
        )

    def test_dummy_backend_init_no_default_app_config(self):
        import getpaid.backends.dummy

        source = inspect.getsource(getpaid.backends.dummy)
        assert 'default_app_config' not in source


class TestNoTypingExtensions:
    """TypedDict is in stdlib since Python 3.8; typing_extensions is unnecessary."""

    def test_types_module_uses_stdlib_typeddict(self):
        source_file = importlib.import_module('getpaid.types').__file__
        source = pathlib.Path(source_file).read_text()
        assert 'typing_extensions' not in source
        assert 'from typing import' in source or 'typing.TypedDict' in source


class TestNoClasspropertyFallback:
    """classproperty is in django.utils.functional since Django 3.1."""

    def test_types_module_no_try_except_classproperty(self):
        source_file = importlib.import_module('getpaid.types').__file__
        source = pathlib.Path(source_file).read_text()
        assert 'django.utils.decorators' not in source
        assert 'from django.utils.functional import classproperty' in source


class TestDefaultAutoField:
    """AppConfig classes should set default_auto_field."""

    def test_getpaid_appconfig_has_default_auto_field(self):
        from getpaid.apps import GetpaidConfig

        assert hasattr(GetpaidConfig, 'default_auto_field')
        assert GetpaidConfig.default_auto_field is not None

    def test_dummy_appconfig_has_default_auto_field(self):
        from getpaid.backends.dummy.apps import DummyPluginAppConfig

        assert hasattr(DummyPluginAppConfig, 'default_auto_field')
        assert DummyPluginAppConfig.default_auto_field is not None


class TestNoPhantomDependencies:
    """Verify no imports of declared-but-unused dependencies in library code."""

    def _collect_library_sources(self):
        """Collect all Python source files in getpaid/ package."""
        import pathlib

        getpaid_root = pathlib.Path(
            importlib.import_module('getpaid').__file__
        ).parent
        return list(getpaid_root.rglob('*.py'))

    def test_no_pendulum_import(self):
        """pendulum should not be imported anywhere in getpaid/."""
        for path in self._collect_library_sources():
            source = path.read_text()
            assert 'import pendulum' not in source, (
                f'Found pendulum import in {path}'
            )

    def test_no_django_model_utils_import(self):
        """django-model-utils should not be imported anywhere in getpaid/."""
        for path in self._collect_library_sources():
            source = path.read_text()
            assert 'import model_utils' not in source, (
                f'Found model_utils import in {path}'
            )
            assert 'from model_utils' not in source, (
                f'Found model_utils import in {path}'
            )

    def test_no_requests_import_in_library(self):
        """requests should not be imported in the core library or dummy backend."""
        for path in self._collect_library_sources():
            source = path.read_text()
            # Check for direct 'import requests' or 'from requests'
            for line in source.splitlines():
                stripped = line.strip()
                if stripped.startswith('#'):
                    continue
                assert not (
                    stripped.startswith('import requests')
                    or stripped.startswith('from requests')
                ), f'Found requests import in {path}: {stripped}'

    def test_pyproject_declares_django_dependency(self):
        """pyproject.toml should declare Django as a dependency."""
        import pathlib

        pyproject_path = (
            pathlib.Path(
                importlib.import_module('getpaid').__file__
            ).parent.parent
            / 'pyproject.toml'
        )
        source = pyproject_path.read_text()
        # Should have Django in dependencies (case insensitive check)
        assert 'django' in source.lower() or 'Django' in source

    def test_pyproject_no_pendulum_dependency(self):
        """pyproject.toml should not declare pendulum as a dependency."""
        import pathlib

        pyproject_path = (
            pathlib.Path(
                importlib.import_module('getpaid').__file__
            ).parent.parent
            / 'pyproject.toml'
        )
        source = pyproject_path.read_text()
        # Should not have pendulum in [tool.poetry.dependencies]
        in_deps = False
        for line in source.splitlines():
            if '[tool.poetry.dependencies]' in line:
                in_deps = True
                continue
            if in_deps and line.startswith('['):
                break
            if in_deps and 'pendulum' in line.lower():
                assert False, f'pendulum still in dependencies: {line}'

    def test_pyproject_no_django_model_utils_dependency(self):
        """pyproject.toml should not declare django-model-utils as a dependency."""
        import pathlib

        pyproject_path = (
            pathlib.Path(
                importlib.import_module('getpaid').__file__
            ).parent.parent
            / 'pyproject.toml'
        )
        source = pyproject_path.read_text()
        in_deps = False
        for line in source.splitlines():
            if '[tool.poetry.dependencies]' in line:
                in_deps = True
                continue
            if in_deps and line.startswith('['):
                break
            if in_deps and 'django-model-utils' in line.lower():
                assert False, (
                    f'django-model-utils still in dependencies: {line}'
                )

    def test_pyproject_no_typing_extensions_dependency(self):
        """pyproject.toml should not declare typing-extensions as a dependency."""
        import pathlib

        pyproject_path = (
            pathlib.Path(
                importlib.import_module('getpaid').__file__
            ).parent.parent
            / 'pyproject.toml'
        )
        source = pyproject_path.read_text()
        in_deps = False
        for line in source.splitlines():
            if '[tool.poetry.dependencies]' in line:
                in_deps = True
                continue
            if in_deps and line.startswith('['):
                break
            if in_deps and 'typing-extensions' in line.lower():
                assert False, f'typing-extensions still in dependencies: {line}'


class TestStatusCompat:
    """getpaid.status should still re-export symbols for backward compat."""

    def test_status_exports_payment_status(self):
        from getpaid.status import PaymentStatus
        from getpaid.types import PaymentStatus as PS

        assert PaymentStatus is PS

    def test_status_exports_fraud_status(self):
        from getpaid.status import FraudStatus
        from getpaid.types import FraudStatus as FS

        assert FraudStatus is FS



from django.urls import path

from . import views

app_name = 'paywall'

urlpatterns = [
    path('fake_gateway/', views.authorization_view, name='gateway'),
    path('fake_gateway_api/<uuid:pk>/', views.get_status, name='get_status'),
    path('fake_gateway_api/', views.rest_register_payment, name='api_register'),
    path('fake_gateway_api/operate/', views.rest_operation, name='api_operate'),
]



import swapper
from django import forms
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from getpaid.validators import run_getpaid_validators

Order = swapper.load_model('getpaid', 'Order')


class PaymentMethodForm(forms.ModelForm):
    """
    Usable example.
    Displays all available payments backends as choice list.
    """

    order = forms.ModelChoiceField(
        widget=forms.HiddenInput, queryset=Order.objects.all()
    )

    class Meta:
        model = swapper.load_model('getpaid', 'Payment')
        fields = [
            'order',
            'amount_required',
            'description',
            'currency',
            'backend',
        ]
        widgets = {
            'amount_required': forms.HiddenInput,
            'description': forms.HiddenInput,
            'currency': forms.HiddenInput,
        }

    def __init__(self, *args, **kwargs):
        from .registry import registry

        super().__init__(*args, **kwargs)
        order = self.initial.get('order')
        currency = getattr(order, 'currency', None) or self.data.get('currency')
        if order is not None:
            self.initial['amount_required'] = order.get_total_amount()
            self.initial['description'] = order.get_description()
        backends = registry.get_choices(currency)
        params = dict(
            choices=backends,
            initial=backends[0][0] if len(backends) == 1 else '',
            label=_('Payment backend'),
            widget=forms.RadioSelect,
        )
        hide_lonely = getattr(settings, 'GETPAID', {}).get(
            'HIDE_LONELY_PLUGIN', False
        )
        if hide_lonely and len(backends) == 1:
            params['initial'] = backends[0][0]
            params['widget'] = forms.HiddenInput

        self.fields['backend'] = forms.ChoiceField(**params)

    def clean_order(self):
        if hasattr(self.cleaned_data['order'], 'is_ready_for_payment'):
            if not self.cleaned_data['order'].is_ready_for_payment():
                raise forms.ValidationError(
                    _('Order is not ready for payment.')
                )
        return self.cleaned_data['order']

    def _apply_order_defaults(self, cleaned_data):
        order = cleaned_data.get('order')
        if order is None:
            return cleaned_data

        cleaned_data['amount_required'] = order.get_total_amount()
        cleaned_data['description'] = order.get_description()
        cleaned_data['currency'] = order.get_currency()
        return cleaned_data

    def clean(self):
        cleaned_data = super().clean()
        cleaned_data = self._apply_order_defaults(cleaned_data)
        return run_getpaid_validators(cleaned_data)



from decimal import Decimal

from django.db import models

from getpaid.abstracts import AbstractOrder


class Order(AbstractOrder):
    description = models.CharField(max_length=100, default='Test order')
    total = models.DecimalField(decimal_places=2, max_digits=8, default='10.00')
    currency = models.CharField(max_length=3, default='EUR')

    def get_absolute_url(self):
        return '/order/'

    def get_total_amount(self):
        return Decimal(self.total)

    def get_buyer_info(self):
        return {'email': 'test@example.com'}

    def get_description(self):
        return self.description
